# -*- coding: utf-8 -*-
"""
rex bias correction utilities.
"""
import os
from concurrent.futures import ProcessPoolExecutor, as_completed
import logging
import numpy as np
import scipy

logger = logging.getLogger(__name__)


def sample_q_linear(n_samples):
    """Sample quantiles from 0 to 1 inclusive linearly with even spacing

    Parameters
    ----------
    n_samples : int
        Number of points to sample between 0 and 1

    Returns
    -------
    quantiles : np.ndarray
        1D array of evenly spaced samples from 0 to 1
    """
    quantiles = np.linspace(0, 1, n_samples)
    return quantiles


def sample_q_log(n_samples, log_base):
    """Sample quantiles from 0 to 1 while concentrating samples near quantile=0

    Parameters
    ----------
    n_samples : int
        Number of points to sample between 0 and 1
    log_base : int | float
        Log base value. A higher value will concentrate more samples at the
        extreme sides of the distribution.

    Returns
    -------
    quantiles : np.ndarray
        1D array of log-spaced samples from 0 to 1
    """
    quantiles = np.logspace(0, 1, n_samples, base=log_base)
    quantiles = (quantiles - 1) / (log_base - 1)
    return quantiles


def sample_q_invlog(n_samples, log_base):
    """Sample quantiles from 0 to 1 while concentrating samples near quantile=1

    Parameters
    ----------
    n_samples : int
        Number of points to sample between 0 and 1
    log_base : int | float
        Log base value. A higher value will concentrate more samples at the
        extreme sides of the distribution.

    Returns
    -------
    quantiles : np.ndarray
        1D array of log-spaced samples from 0 to 1
    """
    quantiles = np.logspace(0, 1, n_samples, base=log_base)
    quantiles = (quantiles - 1) / (log_base - 1)
    quantiles = np.array(sorted(1 - quantiles))
    return quantiles


class QuantileDeltaMapping:
    """Class for quantile delta mapping based on the method from
    Cannon et al., 2015

    Note that this is a utility class for implementing QDM and should not be
    requested directly as a method in the reV/rex bias correction table input

    Cannon, A. J., Sobie, S. R. & Murdock, T. Q. Bias Correction of GCM
    Precipitation by Quantile Mapping: How Well Do Methods Preserve Changes in
    Quantiles and Extremes? Journal of Climate 28, 6938–6959 (2015).
    """

    def __init__(self, params_oh, params_mh, params_mf, dist='empirical',
                 relative=True, sampling='linear', log_base=10):
        """
        Parameters
        ----------
        params_oh : np.ndarray
            2D array of **observed historical** distribution parameters created
            from a multi-year set of data where the shape is (space, N). This
            can be the output of a parametric distribution fit like
            ``scipy.stats.weibull_min.fit()`` where N is the number of
            parameters for that distribution, or this can define the x-values
            of N points from an empirical CDF that will be linearly
            interpolated between. If this is an empirical CDF, this must
            include the 0th and 100th percentile values and have even
            percentile spacing between values.
        params_mh : np.ndarray
            Same requirements as params_oh. This input arg is for the **modeled
            historical distribution**.
        params_mf : np.ndarray | None
            Same requirements as params_oh. This input arg is for the **modeled
            future distribution**. If this is None, this defaults to params_mh
            (no future data, just corrected to modeled historical distribution)
        dist : str | np.ndarray
            Probability distribution name to use to model the data which
            determines how the param args are used. This can "empirical" or any
            continuous distribution name from ``scipy.stats``. Can also be a 1D
            array of dist inputs if being used from reV, but they must all be
            the same option.
        relative : bool | np.ndarray
            Flag to preserve relative rather than absolute changes in
            quantiles. relative=False (default) will multiply by the change in
            quantiles while relative=True will add. See Equations 4-6 from
            Cannon et al., 2015 for more details. Can also be a 1D array of
            dist inputs if being used from reV, but they must all be the same
            option.
        sampling : str | np.ndarray
            If dist="empirical", this is an option for how the quantiles were
            sampled to produce the params inputs, e.g., how to sample the
            y-axis of the distribution (see sampling functions in
            ``rex.utilities.bc_utils``). "linear" will do even spacing, "log"
            will concentrate samples near quantile=0, and "invlog" will
            concentrate samples near quantile=1. Can also be a 1D array of dist
            inputs if being used from reV, but they must all be the same
            option.
        log_base : int | float | np.ndarray
            Log base value if sampling is "log" or "invlog". A higher value
            will concentrate more samples at the extreme sides of the
            distribution. Can also be a 1D array of dist inputs if being used
            from reV, but they must all be the same option.
        """

        self.params_oh = params_oh
        self.params_mh = params_mh
        self.params_mf = params_mf if params_mf is not None else params_mh
        self.relative = bool(self._clean_kwarg(relative))
        self.dist_name = str(self._clean_kwarg(dist)).casefold()
        self.sampling = str(self._clean_kwarg(sampling)).casefold()
        self.log_base = float(self._clean_kwarg(log_base))
        self.scipy_dist = None

        if self.dist_name != 'empirical':
            self.scipy_dist = getattr(scipy.stats, self.dist_name, None)
            if self.scipy_dist is None:
                msg = ('Could not get requested distribution "{}" from '
                       '``scipy.stats``. Please double check your spelling '
                       'and select "empirical" or one of the continuous '
                       'distribution options from here: '
                       'https://docs.scipy.org/doc/scipy/reference/stats.html'
                       .format(self.dist_name))
                logger.error(msg)
                raise KeyError(msg)

    @staticmethod
    def _clean_kwarg(inp):
        """Clean any kwargs inputs (e.g., dist, relative) that might be
        provided as an array and must be collapsed into a single string or
        boolean value"""
        unique = np.unique(inp)
        msg = ('_QuantileDeltaMapping kwargs must have only one unique input '
               'even if being called with arrays as part of reV but found: {}'
               .format(unique))
        assert len(unique) == 1, msg

        while isinstance(inp, np.ndarray):
            inp = inp[0]

        return inp

    @staticmethod
    def _clean_params(params, arr_shape, scipy_dist):
        """Verify and clean 2D parameter arrays for passing into empirical
        distribution or scipy continuous distribution functions.

        Parameters
        ----------
        params : np.ndarray
            Input params shape should be (space, N) where N is the number of
            parameters for the distribution.
        arr_shape : tuple
            Array shape should be (time, space).
        scipy_dist : scipy.stats.rv_continuous | None
            Any continuous distribution class from ``scipy.stats`` or None if
            using an empirical distribution (taken from attribute
            ``QuantileDeltaMapping.scipy_dist``)

        Returns
        -------
        params : np.ndarray | list
            If a scipy continuous dist is set, this output will be params
            unpacked along axis=1 into a list so that the list entries
            represent the scipy distribution parameters
            (e.g., shape, scale, loc) and each list entry is of shape (space,)
        """

        msg = f'params must be 2D array but received {type(params)}'
        assert hasattr(params, 'shape'), msg

        if len(params.shape) == 1:
            params = np.expand_dims(params, 0)

        msg = (f'params must be 2D array of shape ({arr_shape[1]}, N) '
               f'but received shape {params.shape}')
        assert len(params.shape) == 2, msg
        assert params.shape[0] == arr_shape[1], msg

        if scipy_dist is not None:
            params = [params[:, i] for i in range(params.shape[1])]

        return params

    @staticmethod
    def _get_quantiles(n_samples, sampling, log_base):
        """If dist='empirical', this will get the quantile values for the CDF
        x-values specified in the input params"""

        if sampling == 'linear':
            quantiles = sample_q_linear(n_samples)
        elif sampling == 'log':
            quantiles = sample_q_log(n_samples, log_base)
        elif sampling == 'invlog':
            quantiles = sample_q_invlog(n_samples, log_base)
        else:
            msg = ('sampling option must be linear, log, or invlog, but '
                   'received: {}'.format(sampling))
            logger.error(msg)
            raise KeyError(msg)

        return quantiles

    @classmethod
    def cdf(cls, x, params, scipy_dist, sampling, log_base):
        """Run the CDF function e.g., convert physical variable to quantile"""

        if scipy_dist is None:
            p = np.zeros_like(x)
            for idx in range(x.shape[1]):
                xp = params[idx, :]
                fp = cls._get_quantiles(len(xp), sampling, log_base)
                p[:, idx] = np.interp(x[:, idx], xp, fp)
        else:
            p = scipy_dist.cdf(x, *params)

        return p

    @classmethod
    def ppf(cls, p, params, scipy_dist, sampling, log_base):
        """Run the inverse CDF function (percent point function) e.g., convert
        quantile to physical variable"""

        if scipy_dist is None:
            x = np.zeros_like(p)
            for idx in range(p.shape[1]):
                fp = params[idx, :]
                xp = cls._get_quantiles(len(fp), sampling, log_base)
                x[:, idx] = np.interp(p[:, idx], xp, fp)
        else:
            x = scipy_dist.ppf(p, *params)

        return x

    @classmethod
    def run_qdm(cls, arr, params_oh, params_mh, params_mf,
                scipy_dist, relative, sampling, log_base):
        """Run the actual QDM operation from args without initializing the
        ``QuantileDeltaMapping`` object

        Parameters
        ----------
        arr : np.ndarray
            2D array of values in shape (time, space)
        params_oh : np.ndarray
            2D array of **observed historical** distribution parameters created
            from a multi-year set of data where the shape is (space, N). This
            can be the output of a parametric distribution fit like
            ``scipy.stats.weibull_min.fit()`` where N is the number of
            parameters for that distribution, or this can define the x-values
            of N points from an empirical CDF that will be linearly
            interpolated between. If this is an empirical CDF, this must
            include the 0th and 100th percentile values and have even
            percentile spacing between values.
        params_mh : np.ndarray
            Same requirements as params_oh. This input arg is for the **modeled
            historical distribution**.
        params_mf : np.ndarray
            Same requirements as params_oh. This input arg is for the **modeled
            future distribution**.
        scipy_dist : scipy.stats.rv_continuous | None
            Any continuous distribution class from ``scipy.stats`` or None if
            using an empirical distribution (taken from attribute
            ``QuantileDeltaMapping.scipy_dist``)
        relative : bool | np.ndarray
            Flag to preserve relative rather than absolute changes in
            quantiles. relative=False (default) will multiply by the change in
            quantiles while relative=True will add. See Equations 4-6 from
            Cannon et al., 2015 for more details. Can also be a 1D array of
            dist inputs if being used from reV, but they must all be the same
            option.
        sampling : str | np.ndarray
            If dist="empirical", this is an option for how the quantiles were
            sampled to produce the params inputs, e.g., how to sample the
            y-axis of the distribution (see sampling functions in
            ``rex.utilities.bc_utils``). "linear" will do even spacing, "log"
            will concentrate samples near quantile=0, and "invlog" will
            concentrate samples near quantile=1. Can also be a 1D array of dist
            inputs if being used from reV, but they must all be the same
            option.
        log_base : int | float | np.ndarray
            Log base value if sampling is "log" or "invlog". A higher value
            will concentrate more samples at the extreme sides of the
            distribution. Can also be a 1D array of dist inputs if being used
            from reV, but they must all be the same option.

        Returns
        -------
        arr : np.ndarray
            Bias corrected copy of the input array with same shape.
        """

        params_oh = cls._clean_params(params_oh, arr.shape, scipy_dist)
        params_mh = cls._clean_params(params_mh, arr.shape, scipy_dist)
        params_mf = cls._clean_params(params_mf, arr.shape, scipy_dist)

        # Equation references are from Section 3 of Cannon et al 2015:
        # Cannon, A. J., Sobie, S. R. & Murdock, T. Q. Bias Correction of GCM
        # Precipitation by Quantile Mapping: How Well Do Methods Preserve
        # Changes in Quantiles and Extremes? Journal of Climate 28, 6938–6959
        # (2015).

        logger.debug('Computing CDF on modeled future data')
        # Eq.3: Tau_m_p = F_m_p(x_m_p)
        q_mf = cls.cdf(arr, params_mf, scipy_dist, sampling, log_base)

        logger.debug('Computing PPF on observed historical data')
        # Eq.5: x^_o:m_h:p = F-1_o_h(Tau_m_p)
        x_oh = cls.ppf(q_mf, params_oh, scipy_dist, sampling, log_base)

        logger.debug('Computing PPF on modeled historical data')
        # Eq.4 denom: F-1_m_h(Tau_m_p)
        x_mh_mf = cls.ppf(q_mf, params_mh, scipy_dist, sampling, log_base)

        logger.debug('Finished computing distributions.')

        if relative:
            x_mh_mf[x_mh_mf == 0] = 0.001  # arbitrary limit to prevent div 0
            delta = arr / x_mh_mf  # Eq.4: x_m_p / F-1_m_h(Tau_m_p)
            arr_bc = x_oh * delta  # Eq.6: x^_m_p = x^_o:m_h:p * delta
        else:
            delta = arr - x_mh_mf  # Eq.4: x_m_p - F-1_m_h(Tau_m_p)
            arr_bc = x_oh + delta  # Eq.6: x^_m_p = x^_o:m_h:p + delta

        return arr_bc

    def __call__(self, arr, max_workers=1):
        """Run the QDM function to bias correct an array

        Parameters
        ----------
        arr : np.ndarray
            2D array of values in shape (time, space)
        max_workers : int, None
            Number of parallel workers to use in QDM bias correction. 1 will
            run in serial (default), None will use all available cores.

        Returns
        -------
        arr : np.ndarray
            Bias corrected copy of the input array with same shape.
        """

        if len(arr.shape) == 1:
            arr = np.expand_dims(arr, 1)

        if max_workers == 1:
            arr_bc = self.run_qdm(arr, self.params_oh, self.params_mh,
                                  self.params_mf, self.scipy_dist,
                                  self.relative, self.sampling, self.log_base)
        else:
            max_workers = max_workers or os.cpu_count()
            sslices = np.array_split(np.arange(arr.shape[1]), arr.shape[1])
            sslices = [slice(idx[0], idx[-1] + 1) for idx in sslices]
            arr_bc = arr.copy()
            futures = {}
            with ProcessPoolExecutor(max_workers=max_workers) as exe:
                for idx in range(arr.shape[1]):
                    idx = slice(idx, idx + 1)
                    fut = exe.submit(self.run_qdm, arr[:, idx],
                                     self.params_oh[idx],
                                     self.params_mh[idx],
                                     self.params_mf[idx], self.scipy_dist,
                                     self.relative, self.sampling,
                                     self.log_base)
                    futures[fut] = idx
                for future in as_completed(futures):
                    idx = futures[future]
                    arr_bc[:, idx] = future.result()

        msg = ('Input shape {} does not match QDM bias corrected output '
               'shape {}!'.format(arr.shape, arr_bc.shape))
        assert arr.shape == arr_bc.shape, msg

        return arr_bc



# -*- coding: utf-8 -*-
"""
The REsource eXtraction tool (rex)
"""
from __future__ import print_function, division, absolute_import
import logging
if not logging.getLogger().handlers:
    logging.getLogger().addHandler(logging.NullHandler())
import os

from rex.joint_pd import JointPD
from rex.multi_file_resource import (MultiFileNSRDB, MultiFileResource,
                                     MultiFileWTK)
from rex.multi_time_resource import (MultiTimeNSRDB, MultiTimeResource,
                                     MultiTimeWaveResource,
                                     MultiTimeWindResource)
from rex.multi_year_resource import (MultiYearNSRDB, MultiYearResource,
                                     MultiYearWaveResource,
                                     MultiYearWindResource)
from rex.multi_res_resource import MultiResolutionResource
from rex.rechunk_h5 import (ArrayChunkSize, TimeseriesChunkSize, CombineH5,
                            RechunkH5, get_dataset_attributes)
from rex.renewable_resource import (NSRDB, WaveResource, WindResource,
                                    GeothermalResource)
from rex.resource_extraction import (ResourceX, MultiFileResourceX,
                                     MultiTimeResourceX, MultiYearResourceX,
                                     SolarX, NSRDBX, MultiFileNSRDBX,
                                     MultiTimeNSRDBX, MultiYearNSRDBX,
                                     WindX, MultiFileWindX,
                                     MultiTimeWindX, MultiYearWindX,
                                     WaveX, MultiTimeWaveX, MultiYearWaveX)
from rex.temporal_stats import TemporalStats
from rex.utilities import (SpawnProcessPool, SLURM, init_logger, init_mult,
                           setup_logger, log_mem, log_versions, LOGGERS,
                           SolarPosition, safe_json_load, jsonify_dict,
                           parse_year, check_res_file, parse_table,
                           check_eval_str, to_records_array)
from rex.resource import Resource
from rex.outputs import Outputs
from rex.version import __version__

__author__ = """Michael Rossol"""
__email__ = "michael.rossol@nrel.gov"

REXDIR = os.path.dirname(os.path.realpath(__file__))
TESTDATADIR = os.path.join(os.path.dirname(REXDIR), 'tests', 'data')
TREEDIR = os.path.join(os.path.dirname(REXDIR), 'bin', 'trees')



# -*- coding: utf-8 -*-
"""
Classes to handle resource data stored over multiple files
"""
import os
from fnmatch import fnmatch
from glob import glob

import numpy as np
import pandas as pd

from rex.renewable_resource import (
    NSRDB,
    SolarResource,
    WaveResource,
    WindResource,
)
from rex.resource import Resource
from rex.utilities.exceptions import FileInputError
from rex.utilities.parse_keys import parse_keys, parse_slice


class MultiTimeH5:
    """
    Class to handle h5 Resources stored over multiple temporal files
    """

    def __init__(self, h5_path, res_cls=Resource, hsds=False, hsds_kwargs=None,
                 **res_cls_kwargs):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be an
            explicit list of multi time files.
        res_cls : obj
            Resource class to use to open and access resource data
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        res_cls_kwargs : dict, optional
            Kwargs for `res_cls`
        """
        self.h5_path = h5_path
        self._file_paths = self._get_file_paths(h5_path, hsds=hsds,
                                                hsds_kwargs=hsds_kwargs)
        res_cls_kwargs.update({'hsds': hsds})
        self._h5_map = self._map_file_instances(self._file_paths,
                                                res_cls=res_cls,
                                                **res_cls_kwargs)

        self._datasets = None
        self._shape = None
        self._time_index = None
        self._time_slice_map = []
        self._i = 0

    def __repr__(self):
        msg = ("{} for {}:\n Contains data from {} files"
               .format(self.__class__.__name__, self.h5_path, len(self)))
        return msg

    def __getitem__(self, file):
        fn_fp_map = {os.path.basename(fp): fp for fp in self._file_paths}
        if file in self._h5_map['fp'].values:
            iloc = np.where(self._h5_map['fp'] == file)[0][0]
            h5 = self._h5_map.at[iloc, 'h5']
        elif file in fn_fp_map:
            iloc = np.where(self._h5_map['fp'] == fn_fp_map[file])[0][0]
            h5 = self._h5_map.at[iloc, 'h5']
        else:
            raise ValueError('{} is invalid, must be one of: {}'
                             .format(file, self._file_paths))

        return h5

    @property
    def attrs(self):
        """
        Global .h5 file attributes sourced from first .h5 file

        Returns
        -------
        attrs : dict
        """
        attrs = dict(self.h5.attrs)

        return attrs

    @property
    def files(self):
        """
        Available file paths

        Returns
        -------
        list
        """
        return sorted(self._file_paths)

    @property
    def h5_files(self):
        """
        .h5 files data is being sourced from

        Returns
        -------
        list
        """
        return sorted(self._h5_map['fp'])

    @property
    def h5(self):
        """
        open h5 file handler for a single .h5 file

        Returns
        -------
        h5py.File
        """
        return self._h5_map['h5'].values[0]

    @property
    def datasets(self):
        """
        Available datasets

        Returns
        -------
        list
        """
        if self._datasets is None:
            all_dsets = self._h5_map['dsets'].values.tolist()
            dsets = [d for sub in all_dsets for d in sub]
            self._datasets = list(set(dsets))

        return self._datasets

    @property
    def resource_datasets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        res_dsets = [ds for ds in self.datasets
                     if ds not in ['meta', 'time_index', 'coordinates']]

        return res_dsets

    @property
    def shape(self):
        """
        Dataset shape (time, sites)

        Returns
        -------
        tuple
        """
        if self._shape is None:
            self._shape = (len(self.time_index), self.h5.shape[1])

        return self._shape

    @property
    def time_index(self):
        """
        Multi-year datetime index

        Returns
        -------
        pandas.DatatimeIndex
        """
        if self._time_index is None:
            time_slice_map = []
            for _, row in self._h5_map.iterrows():
                h5 = row['h5']
                fp = row['fp']
                ti = h5.time_index
                time_slice_map.append(np.full(len(ti), os.path.basename(fp)))
                if self._time_index is None:
                    self._time_index = ti
                else:
                    self._time_index = self._time_index.append(ti)

            if len(self._time_index) != len(np.unique(self._time_index)):
                unique, duplicates = np.unique(self._time_index,
                                               return_counts=True)
                duplicates = np.where(duplicates > 1)[0]
                duplicates = unique[duplicates]
                msg = ('The combined time_index has {} duplicate values:\n{}'
                       .format(len(duplicates), duplicates))
                raise RuntimeError(msg)

            self._time_slice_map = np.concatenate(time_slice_map, axis=0)

        return self._time_index

    @staticmethod
    def _get_hsds_file_paths(h5_path, hsds_kwargs=None):
        """
        Get a list of h5 filepaths matching the h5_path specification from HSDS

        Parameters
        ----------
        h5_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes.
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None

        Returns
        -------
        file_paths : list
            List of filepaths for this handler to handle.
        """
        import h5pyd

        if hsds_kwargs is None:
            hsds_kwargs = {}

        hsds_dir = os.path.dirname(h5_path)
        fn = os.path.basename(h5_path)

        if '*' in hsds_dir:
            msg = ('HSDS path specifications cannot handle wildcards in the '
                   'directory name! The directory must be explicit but the '
                   'filename can have wildcards. This HSDS h5_path input '
                   'cannot be used: {}'.format(h5_path))
            raise FileNotFoundError(msg)

        if not fn:
            msg = ('h5_path must be a unix shell style pattern with '
                   'wildcard * in order to find files, but received '
                   'directory specification: {}'.format(h5_path))
            raise FileInputError(msg)

        with h5pyd.Folder(hsds_dir + '/', **hsds_kwargs) as f:
            file_paths = [f'{hsds_dir}/{fn}' for fn in f
                          if fnmatch(f'{hsds_dir}/{fn}', h5_path)]

        return file_paths

    @classmethod
    def _get_file_paths(cls, h5_path, hsds=False, hsds_kwargs=None):
        """
        Get a file list based on the h5_path specification.

        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be an
            explicit list of multi time files.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None

        Returns
        -------
        file_paths : list
            List of filepaths for this handler to handle.
        """

        if hsds:
            file_paths = cls._get_hsds_file_paths(h5_path,
                                                  hsds_kwargs=hsds_kwargs)
        elif isinstance(h5_path, (list, tuple)):
            for fp in h5_path:
                msg = 'Does not exist: {}'.format(fp)
                assert os.path.exists(fp), msg
            file_paths = h5_path
        elif os.path.isdir(h5_path):
            msg = ('h5_path must be a unix shell style pattern with '
                   'wildcard * in order to find files, but received '
                   'directory specification: {}'.format(h5_path))
            raise FileInputError(msg)
        elif isinstance(h5_path, str):
            file_paths = glob(h5_path)

        if not any(file_paths):
            msg = ('Could not find any file paths with pattern: {}'
                   .format(h5_path))
            raise FileInputError(msg)

        return file_paths

    @staticmethod
    def _map_file_instances(file_paths, res_cls=Resource, **res_cls_kwargs):
        """
        Open all .h5 files and map the open h5py instances to the
        associated file paths

        Parameters
        ----------
        file_paths : list
            List of filepaths for this handler to handle.

        Returns
        -------
        h5_map : pd.DataFrame
            DataFrame mapping file paths to open resource instances and
            datasets per file (columns: fp, h5, and dsets)
        """
        h5_map = pd.DataFrame({'fp': file_paths, 'h5': None,
                               't0': None, 't1': None})
        for i, f_path in enumerate(file_paths):
            h5_map.at[i, 'h5'] = res_cls(f_path, **res_cls_kwargs)
            h5_map.at[i, 't0'] = h5_map.at[i, 'h5'].time_index.values[0]
            h5_map.at[i, 't1'] = h5_map.at[i, 'h5'].time_index.values[1]

        h5_map['dsets'] = [h5.dsets for h5 in h5_map['h5'].values]
        h5_map = h5_map.sort_values('t0').reset_index(drop=True)

        return h5_map

    @staticmethod
    def _check_time_slice(time_slice):
        """
        Check to see if time positions can be represented as a slice

        Parameters
        ----------
        time_slice : ndarray | list
            List of temporal positions

        Returns
        -------
        time_slice : ndarray | list | slice
            Slice covering range of positions to extract if possible
        """
        s = time_slice[0]
        e = time_slice[-1] + 1
        if (e - s) == len(time_slice):
            time_slice = slice(s, e, None)

        return time_slice

    def _map_time_slice(self, time_slice):
        """
        Map timeslices to files

        Parameters
        ----------
        time_slice : int | list | slice
            tuple describing slice of dataset array to extract

        Returns
        -------
        file_times : dict
            Dictionary mapping files to the time_slices to extract
        """
        time_index = self.time_index[time_slice]
        files = self._time_slice_map[time_slice]
        file_times = {}
        for file in np.unique(files):
            ti = self[file].time_index
            file_slice = np.where(ti.isin(time_index))[0]
            file_slice = self._check_time_slice(file_slice)
            file_times[file] = file_slice

        return file_times

    def _get_ds(self, ds_name, ds_slice):
        """
        Extract data from given dataset

        Parameters
        ----------
        ds_name : str
            Variable dataset to be extracted
        ds_slice : int | list | slice
            tuple describing slice of dataset array to extract

        Returns
        -------
        out : ndarray
            ndarray of variable timeseries data
            If unscale, returned in native units else in scaled units
        """
        ds_slice = parse_slice(ds_slice)
        out = []
        time_slice = ds_slice[0]
        if isinstance(time_slice, (int, np.integer)):
            time_step = self.time_index[time_slice]
            file = self._time_slice_map[time_slice]
            time_index = self[file].time_index
            time_slice = np.where(time_step == time_index)[0][0]
            file_slice = (time_slice, ) + ds_slice[1:]
            out = self[file]._get_ds(ds_name, file_slice)
        else:
            file_times = self._map_time_slice(ds_slice[0])
            for file, time_slice in file_times.items():
                file_slice = (time_slice, ) + ds_slice[1:]
                out.append(self[file]._get_ds(ds_name, file_slice))

            out = np.concatenate(out, axis=0)

        return out

    def close(self):
        """
        Close all h5py.File instances
        """
        for f in self._h5_map['h5']:
            f.close()


class MultiTimeResource:
    """
    Class to handle resource data stored temporally accross multiple
    .h5 files

    Examples
    --------
    Extracting the resource's Datetime Index

    >>> path = '$TESTDATADIR/nsrdb/ri_100_nsrdb_*.h5'
    >>> with MultiTimeResource(path) as res:
    >>>     ti = res.time_index
    >>>
    >>> ti
    DatetimeIndex(['2012-01-01 00:00:00', '2012-01-01 00:30:00',
                   '2012-01-01 01:00:00', '2012-01-01 01:30:00',
                   '2012-01-01 02:00:00', '2012-01-01 02:30:00',
                   '2012-01-01 03:00:00', '2012-01-01 03:30:00',
                   '2012-01-01 04:00:00', '2012-01-01 04:30:00',
                   ...
                   '2013-12-31 19:00:00', '2013-12-31 19:30:00',
                   '2013-12-31 20:00:00', '2013-12-31 20:30:00',
                   '2013-12-31 21:00:00', '2013-12-31 21:30:00',
                   '2013-12-31 22:00:00', '2013-12-31 22:30:00',
                   '2013-12-31 23:00:00', '2013-12-31 23:30:00'],
                  dtype='datetime64[ns]', length=35088, freq=None)

    NOTE: time_index covers data from 2012 and 2013

    >>> with MultiTimeResource(path) as res:
    >>>     print(res.h5_files)

    ['/Users/mrossol/Git_Repos/rex/tests/data/nsrdb/ri_100_nsrdb_2012.h5',
     '/Users/mrossol/Git_Repos/rex/tests/data/nsrdb/ri_100_nsrdb_2013.h5']

    Data slicing works the same as with "Resource" except axis 0 now covers
    2012 and 2013

    >>> with MultiTimeResource(path) as res:
    >>>     temperature = res['air_temperature']
    >>>
    >>> temperature
    [[ 4.  5.  5. ...  4.  3.  4.]
     [ 4.  4.  5. ...  4.  3.  4.]
     [ 4.  4.  5. ...  4.  3.  4.]
     ...
     [-1. -1.  0. ... -2. -3. -2.]
     [-1. -1.  0. ... -2. -3. -2.]
     [-1. -1.  0. ... -2. -3. -2.]]
    >>> temperature.shape
    (35088, 100)

    >>> with MultiTimeResource(path) as res:
    >>>     temperature = res['air_temperature', ::100] # every 100th timestep
    >>>
    >>> temperature
    [[ 4.  5.  5. ...  4.  3.  4.]
     [ 1.  1.  2. ...  0.  0.  1.]
     [-2. -1. -1. ... -2. -4. -2.]
     ...
     [-3. -2. -2. ... -3. -4. -3.]
     [ 0.  0.  1. ...  0. -1.  0.]
     [ 3.  3.  3. ...  2.  2.  3.]]
    >>> temperature.shape
    (351, 100)
    """

    def __init__(self, h5_path, unscale=True, str_decode=True,
                 res_cls=Resource, hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be an
            explicit list of multi time files.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        res_cls : obj
            Resource handler to us to open individual .h5 files
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """

        self.h5_path = h5_path
        self._time_index = None
        # Map variables to their .h5 files
        cls_kwargs = {'unscale': unscale, 'str_decode': str_decode,
                      'hsds': hsds, 'hsds_kwargs': hsds_kwargs}
        self._h5 = MultiTimeH5(self.h5_path, res_cls=res_cls, **cls_kwargs)
        self.h5_files = self._h5.h5_files
        self.h5_file = self.h5_files[0]
        self._i = 0

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self.h5_path)
        return msg

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def __len__(self):
        return len(self.h5.time_index)

    def __iter__(self):
        return self

    def __next__(self):
        if self._i >= len(self.datasets):
            self._i = 0
            raise StopIteration

        dset = self.datasets[self._i]
        self._i += 1

        return dset

    def __getitem__(self, keys):
        ds, ds_slice = parse_keys(keys)

        if ds.endswith('time_index'):
            out = self.h5.h5._get_time_index(ds_slice)
        elif ds.endswith('meta'):
            out = self.h5.h5._get_meta(ds, ds_slice)
        elif ds.endswith('coordinates'):
            out = self.h5.h5._get_coords(ds, ds_slice)
        else:
            out = self.h5._get_ds(ds, ds_slice)

        return out

    def __contains__(self, dset):
        return dset in self.datasets

    @property
    def h5(self):
        """
        Open class instance that handles all .h5 files that data is to
        be extracted from

        Returns
        -------
        h5 : MultiTimeH5 | MultiYearH5
        """
        return self._h5

    @property
    def datasets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self.h5.datasets

    @property
    def dsets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self.datasets

    @property
    def resource_datasets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        return self.h5.resource_datasets

    @property
    def res_dsets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        return self.resource_datasets

    @property
    def shape(self):
        """
        Resource shape (timesteps, sites)
        shape = (len(time_index), len(meta))

        Returns
        -------
        shape : tuple
        """
        return self.h5.shape

    @property
    def meta(self):
        """
        Resource meta data DataFrame

        Returns
        -------
        meta : pandas.DataFrame
        """

        return self.h5.h5.meta

    @property
    def time_index(self):
        """
        Resource DatetimeIndex

        Returns
        -------
        time_index : pandas.DatetimeIndex
        """
        return self.h5.time_index

    @property
    def lat_lon(self):
        """
        Extract (latitude, longitude) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.h5.h5.lat_lon

    @property
    def coordinates(self):
        """
        Coordinates: (lat, lon) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.lat_lon

    @property
    def global_attrs(self):
        """
        Global (file) attributes

        Returns
        -------
        global_attrs : dict
        """
        return self.get_attrs()

    @property
    def attrs(self):
        """
        Dictionary of all dataset attributes

        Returns
        -------
        attrs : dict
        """
        return self.h5.h5.attrs

    @property
    def shapes(self):
        """
        Dictionary of all dataset shapes

        Returns
        -------
        shapes : dict
        """
        return self.h5.h5.shapes

    @property
    def dtypes(self):
        """
        Dictionary of all dataset dtypes

        Returns
        -------
        dtypes : dict
        """
        return self.h5.h5.dtypes

    @property
    def chunks(self):
        """
        Dictionary of all dataset chunk sizes

        Returns
        -------
        chunks : dict
        """
        return self.h5.h5.chunks

    @property
    def scale_factors(self):
        """
        Dictionary of all dataset scale factors

        Returns
        -------
        scale_factors : dict
        """
        return self.h5.h5.scale_factors

    @property
    def units(self):
        """
        Dictionary of all dataset units

        Returns
        -------
        units : dict
        """
        return self.h5.h5.units

    def get_attrs(self, dset=None):
        """
        Get h5 attributes either from file or dataset

        Parameters
        ----------
        dset : str
            Dataset to get attributes for, if None get file (global) attributes

        Returns
        -------
        attrs : dict
            Dataset or file attributes
        """
        return self.h5.h5.get_attrs(dset=dset)

    def get_dset_properties(self, dset):
        """
        Get dataset properties (shape, dtype, chunks)

        Parameters
        ----------
        dset : str
            Dataset to get scale factor for

        Returns
        -------
        shape : tuple
            Dataset array shape
        dtype : str
            Dataset array dtype
        chunks : tuple
            Dataset chunk size
        """
        return self.h5.h5.get_dset_properties(dset)

    def get_scale_factor(self, dset):
        """
        Get dataset scale factor

        Parameters
        ----------
        dset : str
            Dataset to get scale factor for

        Returns
        -------
        float
            Dataset scale factor, used to unscale int values to floats
        """
        return self.h5.h5.get_scale_factor(dset)

    def get_units(self, dset):
        """
        Get dataset units

        Parameters
        ----------
        dset : str
            Dataset to get units for

        Returns
        -------
        str
            Dataset units, None if not defined
        """
        return self.h5.h5.get_units(dset)

    def get_meta_arr(self, rec_name, rows=slice(None)):
        """Get a meta array by name (faster than DataFrame extraction).

        Parameters
        ----------
        rec_name : str
            Named record from the meta data to retrieve.
        rows : slice
            Rows of the record to extract.

        Returns
        -------
        meta_arr : np.ndarray
            Extracted array from the meta data record name.
        """
        meta_arr = self.h5.h5.get_meta_arr(rec_name, rows=rows)

        return meta_arr

    def close(self):
        """
        Close h5 instance
        """
        self._h5.close()


class MultiTimeSolarResource:
    """
    Class to handle solar resource data stored temporaly accross multiple .h5
    files
    """

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be an
            explicit list of multi time files.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=SolarResource)


class MultiTimeNSRDB(MultiTimeResource):
    """
    Class to handle NSRDB data stored temporaly accross multiple .h5
    files
    """

    PREFIX = 'nsrdb'

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be an
            explicit list of multi time files.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=NSRDB)


class MultiTimeWindResource(MultiTimeResource):
    """
    Class to handle wind resource data stored temporaly accross multiple .h5
    files
    """

    PREFIX = 'wtk'

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be an
            explicit list of multi time files.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=WindResource)


class MultiTimeWaveResource(MultiTimeResource):
    """
    Class to handle wave resource data stored temporaly accross multiple .h5
    files
    """

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be an
            explicit list of multi time files.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=WaveResource)



# -*- coding: utf-8 -*-
"""
Resource Extraction Tools
"""
import copy
import logging
import os
import pickle
from tempfile import TemporaryDirectory
from warnings import warn

import h5py
import numpy as np
import pandas as pd
from scipy.spatial import cKDTree

from rex.multi_file_resource import (
    MultiFileNSRDB,
    MultiFileResource,
    MultiFileWTK,
)
from rex.multi_time_resource import MultiTimeResource
from rex.multi_year_resource import MultiYearResource
from rex.renewable_resource import (
    NSRDB,
    SolarResource,
    WaveResource,
    WindResource,
)
from rex.resource import Resource, ResourceDataset
from rex.temporal_stats.temporal_stats import TemporalStats
from rex.utilities.exceptions import ResourceValueError, ResourceWarning
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_versions
from rex.utilities.utilities import check_tz, parse_year, res_dist_threshold

# pylint: disable=consider-using-with
TREE_DIR = TemporaryDirectory()
logger = logging.getLogger(__name__)


class ResourceX:
    """
    Resource data extraction tool
    """

    DEFAULT_RES_CLS = Resource

    def __init__(self, res_h5, res_cls=None, tree=None, unscale=True,
                 str_decode=True, group=None, hsds=False, hsds_kwargs=None,
                 log_vers=True):
        """
        Parameters
        ----------
        res_h5 : str
            Path to resource .h5 file of interest
        res_cls : obj, optional
            Resource class to use to open and access resource data,
            by default Resource (default changes for subclasses like NSRDBX)
        tree : str | cKDTree, optional
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates, by default None
        unscale : bool, optional
            Boolean flag to automatically unscale variables on extraction,
            by default True
        str_decode : bool, optional
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
            by default True
        group : str, optional
            Group within .h5 resource file to open, by default None
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        log_vers : bool
            Flag to log rex versions, True by default. Disable this if wrapping
            in a parallel process (logs get very verbose).
        """

        if log_vers:
            log_versions(logger)

        res_cls = self.DEFAULT_RES_CLS if res_cls is None else res_cls
        self._res = res_cls(res_h5, unscale=unscale, str_decode=str_decode,
                            group=group, hsds=hsds, hsds_kwargs=hsds_kwargs)
        self._dist_thresh = None
        self._tree = tree
        self._i = 0

    def __repr__(self):
        msg = "{} extractor for {}".format(self._res.__class__.__name__,
                                           self.resource.h5_file)

        return msg

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def __len__(self):
        return len(self.resource)

    def __getitem__(self, keys):
        return self.resource[keys]

    def __contains__(self, dset):
        return dset in self.datasets

    def __iter__(self):
        return self

    def __next__(self):
        if self._i >= len(self.datasets):
            self._i = 0
            raise StopIteration

        dset = self.datasets[self._i]
        self._i += 1

        return dset

    @property
    def resource(self):
        """
        Open res_cls instance to access res_h5 data

        Returns
        -------
        res_cls : rex.resource.Resource | rex.renewable_resource.*
        """
        return self._res

    @property
    def h5(self):
        """
        Open h5py File instance. If _group is not None return open Group

        Returns
        -------
        h5 : h5py.File | h5py.Group
        """
        return self.resource.h5

    @property
    def datasets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self.resource.datasets

    @property
    def dsets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self.datasets

    @property
    def resource_datasets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        return self.resource.resource_datasets

    @property
    def res_dsets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        return self.resource_datasets

    @property
    def groups(self):
        """
        Groups available

        Returns
        -------
        groups : list
        """
        return self.resource.groups

    @property
    def shape(self):
        """
        Resource shape (timesteps, sites)
        shape = (len(time_index), len(meta))

        Returns
        -------
        shape : tuple
        """
        return self.resource.shape

    @property
    def meta(self):
        """
        Resource meta data DataFrame

        Returns
        -------
        meta : pandas.DataFrame
        """
        return self.resource.meta

    @property
    def time_index(self):
        """
        Resource DatetimeIndex

        Returns
        -------
        time_index : pandas.DatetimeIndex
        """
        return self.resource.time_index

    @property
    def coordinates(self):
        """
        Coordinates: (lat, lon) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.resource.lat_lon

    @property
    def lat_lon(self):
        """
        Extract (latitude, longitude) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.resource.lat_lon

    @property
    def data_version(self):
        """
        Get the version attribute of the data. None if not available.

        Returns
        -------
        version : str | None
        """
        return self.resource.data_version

    @property
    def global_attrs(self):
        """
        Global (file) attributes

        Returns
        -------
        global_attrs : dict
        """
        return self.resource.global_attrs

    @property
    def attrs(self):
        """
        Global (file) attributes

        Returns
        -------
        attrs : dict
        """
        return self.resource.attrs

    @property
    def tree(self):
        """
        Pre-initialized cKDTree on the resource lat, lon coordinates

        Returns
        -------
        tree : cKDTree
        """
        if not isinstance(self._tree, cKDTree):
            self._tree = self._init_tree(self._tree)

        return self._tree

    @property
    def distance_threshold(self):
        """
        Distance threshold, calculated as half of the diagonal between closest
        resource points, with an extra 5% margin

        Returns
        -------
        float
        """
        if self._dist_thresh is None:
            self._dist_thresh = res_dist_threshold(self._res.lat_lon,
                                                   tree=self.tree)

        return self._dist_thresh

    @property
    def countries(self):
        """
        Available Countires

        Returns
        -------
        countries : ndarray
        """
        if 'country' in self.meta:
            countries = self.meta['country'].unique()
        else:
            countries = None

        return countries

    @property
    def states(self):
        """
        Available states

        Returns
        -------
        states : ndarray
        """
        if 'state' in self.meta:
            states = self.meta['state'].unique()
        else:
            states = None

        return states

    @property
    def counties(self):
        """
        Available Counties

        Returns
        -------
        counties : ndarray
        """
        if 'county' in self.meta:
            counties = self.meta['county'].unique()
        else:
            counties = None

        return counties

    @staticmethod
    def _get_tree_file(h5_file):
        """
        Create path to pre-computed tree from h5_file by splitting file name
        at year if available, else replacing the .h5 suffix

        Parameters
        ----------
        h5_file : str
            Path to source .h5 file

        Returns
        -------
        tree_file : str
            Path to pre-comupted tree .pkl file name
        """
        f_name = os.path.basename(h5_file)
        try:
            year = parse_year(f_name)
            tree_file = f_name.split(str(year))[0] + 'tree.pkl'
        except RuntimeError:
            tree_file = f_name.replace('.h5', '_tree.pkl')

        return tree_file

    @staticmethod
    def _load_tree(tree_path):
        """
        Load tree from pickle file

        Parameters
        ----------
        tree_path : str
            Pickle (.pkl, .pickle) file containing precomputed cKDTree

        Returns
        -------
        tree : cKDTree
            Precomputed tree of lat, lon coordinates
        """
        try:
            with open(tree_path, 'rb') as f:
                tree = pickle.load(f)
        except Exception as e:
            logger.warning('Could not extract tree from {}: {}'
                           .format(tree_path, e))
            tree = None

        return tree

    @staticmethod
    def _save_tree(tree, tree_path):
        """
        Save pre-computed Tree to TEMP_DIR as a pickle file

        Parameters
        ----------
        tree : cKDTree
            pre-computed cKDTree
        tree_path : str
            Path to pickle file in TEMP_DIR to save tree too
        """
        try:
            with open(tree_path, 'wb') as f:
                pickle.dump(tree, f)
        except Exception as e:
            logger.warning('Could not save tree to {}: {}'
                           .format(tree_path, e))

    @staticmethod
    def _get_ds_slice(dset, gids):
        """
        Get dataset region slice

        Parameters
        ----------
        dset : str
            Dataset to extract region from
        gids : ndarray | list
            Gids associated with region

        Returns
        -------
        ds_slice : tuple
            ds slice tuple to properly extract region from given dataset
        """
        if dset == 'time_index':
            ds_slice = (slice(None), )
        elif dset in ['coordinates', 'meta']:
            ds_slice = (gids, )
        else:
            ds_slice = (slice(None), gids)

        return ds_slice

    @staticmethod
    def _to_SAM_csv(sam_df, site_meta, out_path, write_time=True):
        """
        Save SAM dataframe to disk and add meta data to header to make
        SAM compliant

        Parameters
        ----------
        sam_df : pandas.DataFrame
            rex SAM DataFrame
        site_meta : pandas.DataFrame
            Site meta data
        out_path : str
            Path to .csv file to save data too
        write_time : bool
            Flag to write the time columns (Year, Month, Day, Hour, Minute)
        """
        if not out_path.endswith('.csv'):
            if os.path.isfile(out_path):
                out_path = os.path.basename(out_path)

            out_path = os.path.join(out_path, "{}.csv".format(sam_df.name))

        if write_time:
            sam_df.to_csv(out_path, index=False)
        else:
            time_cols = ('year', 'month', 'day', 'hour', 'minute')
            cols = [c for c in sam_df if c.lower() not in time_cols]
            sam_df[cols].to_csv(out_path, index=False)

        if 'gid' not in site_meta:
            site_meta.index.name = 'gid'
            site_meta = site_meta.reset_index()

        col_map = {}
        for c in site_meta.columns:
            if c.lower() == 'timezone':
                col_map[c] = 'Time Zone'
            elif c.lower() == 'gid':
                col_map[c] = 'Location ID'
            elif c.islower():
                col_map[c] = c.capitalize()

        site_meta = site_meta.rename(columns=col_map)
        cols = ','.join(site_meta.columns)
        values = site_meta.values[0].astype(str)
        values = ','.join([value.replace(',', '') for value in values])
        values = values.replace('\n', '').replace('\r', '').replace('\t', '')

        with open(out_path, 'r+') as f:
            content = f.read()
            f.seek(0, 0)
            f.write(cols + '\n' + values + '\n' + content)

    def _init_tree(self, tree):
        """
        Inititialize cKDTree of lat, lon coordinates

        Parameters
        ----------
        tree : str | cKDTree | NoneType
            Path to .pgz file containing pre-computed tree
            If None search bin for .pgz file matching h5 file
            else compute tree

        Returns
        -------
        tree : cKDTree
            cKDTree of lat, lon coordinate from wtk .h5 file
        """
        tree_path = self._get_tree_file(self.resource.h5_file)
        if not isinstance(tree, (cKDTree, str, type(None))):
            tree = None
            logger.warning('Precomputed tree must be supplied as a pickle '
                           'file or a cKDTree, not a {}'
                           .format(type(tree)))

        if tree is None:
            if tree_path in os.listdir(TREE_DIR.name):
                tree = os.path.join(TREE_DIR.name, tree_path)

        if isinstance(tree, str):
            tree = self._load_tree(tree)

        if tree is None:
            lat_lon = self.lat_lon
            tree = cKDTree(lat_lon)  # pylint: disable=not-callable
            self._save_tree(tree, os.path.join(TREE_DIR.name, tree_path))

        return tree

    def _check_lat_lon(self, lat_lon):
        """
        Check lat lon coordinates against domain

        Parameters
        ----------
        lat_lon : ndarray
            Either a single (lat, lon) pair or series of (lat, lon) pairs
        """
        lat_min, lat_max = np.sort(self.lat_lon[:, 0])[[0, -1]]
        lon_min, lon_max = np.sort(self.lat_lon[:, 1])[[0, -1]]

        lat = lat_lon[:, 0]
        check = lat < lat_min
        check |= lat > lat_max

        lon = lat_lon[:, 1]
        check |= lon < lon_min
        check |= lon > lon_max

        if any(check):
            bad_coords = lat_lon[check]
            msg = ("Latitude, longitude coordinates ({}) are outsides of the "
                   "resource domain: (({}, {}), ({}, {}))"
                   .format(bad_coords, lat_min, lon_min, lat_max, lon_max))
            raise ResourceValueError(msg)

    def lat_lon_gid(self, lat_lon, check_lat_lon=True):
        """
        Get nearest gid to given (lat, lon) pair or pairs

        Parameters
        ----------
        lat_lon : ndarray
            Either a single (lat, lon) pair or series of (lat, lon) pairs
        check_lat_lon : bool, optional
            Flag to check to make sure the requested lat lons are inside the
            resource grid. This is done by comparing with the bounding box of
            the resource coordinates and by ensuring the nearest neighbor
            distance are below the distance threshold to ensure that requested
            lat, lon coordinates are within the resource grid, by default True

        Returns
        -------
        gids : int | ndarray
            Nearest gid(s) to given (lat, lon) pair(s)
        """
        if not isinstance(lat_lon, np.ndarray):
            lat_lon = np.array(lat_lon, dtype=np.float32)

        if len(lat_lon.shape) == 1:
            lat_lon = np.expand_dims(lat_lon, axis=0)

        dist, gids = self.tree.query(lat_lon)

        if check_lat_lon:
            self._check_lat_lon(lat_lon)
            dist_check = dist > self.distance_threshold
            if np.any(dist_check):
                msg = ("Latitude, longitude coordinates ({}) do not sit within"
                       " resource grid!".format(lat_lon[dist_check]))
                logger.error(msg)
                raise ResourceValueError(msg)

        if len(gids) == 1:
            gids = int(gids[0])

        return gids

    def region_gids(self, region, region_col='state'):
        """
        Get the gids for given region

        Parameters
        ----------
        region : str
            Region to search for
        region_col : str
            Region column to search

        Returns
        -------
        gids : ndarray
            Vector of gids in given region
        """
        gids = self.meta
        gids = gids[gids[region_col] == region].index.values

        return gids

    def box_gids(self, lat_lon_1, lat_lon_2):
        """
        Get gids within bounding lat_lon coordinates

        Parameters
        ----------
        lat_lon_1 : list | tuple
            One corner of the bounding box
        lat_lon_2 : list | tuple
            The other corner of the bounding box

        Returns
        -------
        gids : ndarray
            Gids in bounding box
        """
        self._check_lat_lon(np.vstack((lat_lon_1, lat_lon_2)))
        lat_min, lat_max = sorted([lat_lon_1[0], lat_lon_2[0]])
        lon_min, lon_max = sorted([lat_lon_1[1], lat_lon_2[1]])

        coords = self.lat_lon
        gids = coords[:, 0] >= lat_min
        gids &= coords[:, 0] <= lat_max
        gids &= coords[:, 1] >= lon_min
        gids &= coords[:, 1] <= lon_max

        return np.where(gids)[0]

    def timestep_idx(self, timestep):
        """
        Get the index of the desired timestep

        Parameters
        ----------
        timestep : str
            Timestep of interest

        Returns
        -------
        ts_idx : int
            Time index value
        """
        timestep = check_tz(pd.to_datetime(timestep))
        idx = np.where(self.time_index == timestep)[0][0]

        return idx

    def get_gid_ts(self, ds_name, gid):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        ts : ndarray
            Time-series for given site(s) and dataset
        """
        ts = self[ds_name, :, gid]

        return ts

    def get_gid_df(self, ds_name, gid):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        df : pandas.DataFrame
            Time-series DataFrame for given site(s) and dataset
        """
        index = pd.Index(data=self.time_index, name='time_index')
        if isinstance(gid, (int, np.integer)):
            columns = [gid]
        else:
            columns = gid

        df = pd.DataFrame(self[ds_name, :, gid], columns=columns,
                          index=index)
        df.name = ds_name

        return df

    def get_lat_lon_ts(self, ds_name, lat_lon, check_lat_lon=True):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon : tuple | list
            (lat, lon) coordinate of interest or pairs of coordinates
        check_lat_lon : bool, optional
            Flag to check to make sure the requested lat lons are inside the
            resource grid. This is done by comparing with the bounding box of
            the resource coordinates and by ensuring the nearest neighbor
            distance are below the distance threshold to ensure that requested
            lat, lon coordinates are within the resource grid, by default True

        Return
        ------
        ts : ndarray
            Time-series for given site(s) and dataset
        """
        gid = self.lat_lon_gid(lat_lon, check_lat_lon=check_lat_lon)
        ts = self.get_gid_ts(ds_name, gid)

        return ts

    def get_lat_lon_df(self, ds_name, lat_lon, check_lat_lon=True):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon : tuple
            (lat, lon) coordinate of interest
        check_lat_lon : bool, optional
            Flag to check to make sure the requested lat lons are inside the
            resource grid. This is done by comparing with the bounding box of
            the resource coordinates and by ensuring the nearest neighbor
            distance are below the distance threshold to ensure that requested
            lat, lon coordinates are within the resource grid, by default True

        Return
        ------
        df : pandas.DataFrame
            Time-series DataFrame for given site(s) and dataset
        """
        gid = self.lat_lon_gid(lat_lon, check_lat_lon=check_lat_lon)
        df = self.get_gid_df(ds_name, gid)

        return df

    def get_region_ts(self, ds_name, region, region_col='state'):
        """
        Extract timeseries of of all sites in given region

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        region : str
            Region to search for
        region_col : str
            Region column to search

        Return
        ------
        region_ts : ndarray
            Time-series array of desired dataset for all sites in desired
            region
        """
        gids = self.region_gids(region, region_col=region_col)
        region_ts = self.get_gid_ts(ds_name, gids)

        return region_ts

    def get_region_df(self, ds_name, region, region_col='state'):
        """
        Extract timeseries of of all sites in given region and return as a
        DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        region : str
            Region to extract all pixels for
        region_col : str
            Region column to search

        Return
        ------
        region_df : pandas.DataFrame
            Time-series array of desired dataset for all sites in desired
            region
        """
        gids = self.region_gids(region, region_col=region_col)
        region_df = self.get_gid_df(ds_name, gids)

        return region_df

    def get_box_ts(self, ds_name, lat_lon_1, lat_lon_2):
        """
        Extract timeseries of of all sites in given bounding box

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon_1 : list | tuple
            One corner of the bounding box
        lat_lon_2 : list | tuple
            The other corner of the bounding box

        Return
        ------
        box_ts : ndarray
            Time-series array of desired dataset for all sites in desired
            bounding box
        """
        gids = self.box_gids(lat_lon_1, lat_lon_2)
        box_ts = self.get_gid_ts(ds_name, gids)

        return box_ts

    def get_box_df(self, ds_name, lat_lon_1, lat_lon_2):
        """
        Extract timeseries of of all sites in given bounding box and return as
        a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon_1 : list | tuple
            One corner of the bounding box
        lat_lon_2 : list | tuple
            The other corner of the bounding box

        Return
        ------
        box_df : pandas.DataFrame
            Time-series array of desired dataset for all sites in desired
            bounding box
        """
        gids = self.box_gids(lat_lon_1, lat_lon_2)
        box_df = self.get_gid_df(ds_name, gids)

        return box_df

    def get_SAM_gid(self, gid, out_path=None, write_time=True,
                    extra_meta_data=None, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given resource gid

        Parameters
        ----------
        gid : int | list
            Resource gid(s) of interset
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        write_time : bool
            Flag to write the time columns (Year, Month, Day, Hour, Minute)
        extra_meta_data : dict, optional
            Dictionary that maps the names and values of extra meta
            info. For example, extra_meta_data={'TMY Year': '2020'}
            will add a column 'TMY Year' to the meta data with
            a value of '2020'.
        kwargs : dict
            Internal kwargs for get_SAM_df

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        if isinstance(gid, (int, np.integer)):
            gid = [gid]

        SAM_df = []
        for res_id in gid:
            # pylint: disable=E1111
            df = self.resource.get_SAM_df(res_id, **kwargs)
            SAM_df.append(df)

            if out_path is not None:
                assert out_path.endswith('.csv'), 'out_path must be .csv!'
                i_out_path = out_path
                if len(gid) > 1:
                    tag = '_{}.csv'.format(res_id)
                    i_out_path = i_out_path.replace('.csv', tag)

                site_meta = self['meta', res_id]

                extra_meta_data = extra_meta_data or {}
                for col_name, val in extra_meta_data.items():
                    site_meta[col_name] = val

                if self.data_version is not None:
                    # pylint: disable=unsupported-assignment-operation
                    site_meta['Version'] = self.data_version

                self._to_SAM_csv(df, site_meta, i_out_path,
                                 write_time=write_time)

        if len(SAM_df) == 1:
            SAM_df = SAM_df[0]

        return SAM_df

    def get_SAM_lat_lon(self, lat_lon, check_lat_lon=True, out_path=None,
                        **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given lat_lon

        Parameters
        ----------
        lat_lon : tuple
            (lat, lon) coordinate of interest
        check_lat_lon : bool, optional
            Flag to check to make sure the requested lat lons are inside the
            resource grid. This is done by comparing with the bounding box of
            the resource coordinates and by ensuring the nearest neighbor
            distance are below the distance threshold to ensure that requested
            lat, lon coordinates are within the resource grid, by default True
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        kwargs : dict
            Internal kwargs for get_SAM_df

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        gid = self.lat_lon_gid(lat_lon, check_lat_lon=check_lat_lon)
        SAM_df = self.get_SAM_gid(gid, out_path=out_path, **kwargs)

        return SAM_df

    def get_timestep_map(self, ds_name, timestep, region=None,
                         region_col='state', box=None):
        """
        Extract a map of the given dataset at the given timestep for the
        given region if supplied

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        timestep : str
            Timestep of interest
        region : str, optional
            Region to extract all pixels for, by default None
        region_col : str, optional
            Region column to search, by default 'state'
        box : tuple, optional
            Bounding corners of box to extract pixels for

        Returns
        -------
        ts_map : pandas.DataFrame
            DataFrame of map values
        """
        lat_lons = self.lat_lon
        ts_idx = self.timestep_idx(timestep)
        gids = slice(None)

        if region is not None and box is not None:
            msg = 'Can only process a region OR a set of box corners!'
            raise RuntimeError(msg)

        if region is not None:
            gids = self.region_gids(region, region_col=region_col)
            lat_lons = lat_lons[gids]
        elif box is not None:
            gids = self.box_gids(*box)
            lat_lons = lat_lons[gids]

        ts_map = self[ds_name, ts_idx, gids]
        ts_map = pd.DataFrame({'longitude': lat_lons[:, 1],
                               'latitude': lat_lons[:, 0],
                               ds_name: ts_map})

        return ts_map

    def get_grid_vectors(self, target, meta=None):
        """Get vectors representing pure horizontal/vertical movements in the
        meta data coordinate system. Note that this can break down if a target
        is requested outside of the main grid area.

        Parameters
        ----------
        target : tuple
            Starting coordinate (latitude, longitude) in decimal degrees for
            the bottom left hand corner of the raster grid.
        meta : pd.DataFrame | None
            Optional meta data input with latitude, longitude fields. Default
            is None which extracts self.meta from the resource data.

        Returns
        -------
        gid_target : np.ndarray
            1D array of shape (2,) with (latitude, longitude) corresponding to
            the meta data grid cell closest to the requested target.
        vector_x : np.ndarray
            1D array of shape (2,) with (delta_latitude, delta_longitude)
            corresponding to the vector for pure positive horizontal movement
            in the meta data
        vector_y : np.ndarray
            1D array of shape (2,) with (delta_latitude, delta_longitude)
            corresponding to the vector for pure positive vertical movement in
            the meta data
        close : np.ndarray
            Meta data index values corresponding to the 3x3 box of pixels
            closest to gid_target.
        """

        meta = meta if meta is not None else self.meta

        out_of_bounds = ((target[0] > meta['latitude']).all()
                         | (target[0] < meta['latitude']).all()
                         | (target[1] > meta['longitude']).all()
                         | (target[1] < meta['longitude']).all())
        if out_of_bounds:
            msg = ('Target {} is outside of meta data extent with latitude '
                   'range {} to {} and longitude range {} to {}'
                   .format(target, meta['latitude'].min(),
                           meta['latitude'].max(), meta['longitude'].min(),
                           meta['longitude'].max()))
            raise RuntimeError(msg)

        # find the actual meta data point closest to the target
        dist = ((meta['latitude'] - target[0])**2
                + (meta['longitude'] - target[1])**2)
        target_loc = meta.index.values[np.argmin(dist)]
        gid_target = np.array([meta.at[target_loc, 'latitude'],
                               meta.at[target_loc, 'longitude']])

        # find the 3x3 box of points around the target
        dy = meta['latitude'] - gid_target[0]
        dx = meta['longitude'] - gid_target[1]
        dist = np.sqrt(dx**2 + dy**2)
        close = meta.index.values[np.argsort(dist)[:9]]

        # get the vectors closest to pure horizontal/vertical movement
        theta = np.arctan2(dy.loc[close].values, dx.loc[close].values)
        dx_loc = close[np.argsort(np.abs(theta))[1]]
        dy_loc = close[np.argmin(np.abs(theta - np.pi / 2))]

        # get (delta_latitude, delta_longitude) vectors
        # for pure horizontal/vertical movements
        vector_dx = np.array([dy.loc[dx_loc], dx.loc[dx_loc]])
        vector_dy = np.array([dy.loc[dy_loc], dx.loc[dy_loc]])
        vector_dx[1] = 1e-6 if vector_dx[1] == 0 else vector_dx[1]
        vector_dy[1] = 1e-6 if vector_dy[1] == 0 else vector_dy[1]

        return gid_target, vector_dx, vector_dy, close

    @staticmethod
    def _order_raster_index(raster_index, meta, shape,
                            vec_dy, lat_descending=True):
        """Ensure that the raster index is propertly sorted.

        Parameters
        ----------
        raster_index : np.ndarray
            2D array of meta data index values that form a 2D rectangular grid
        meta : pd.DataFrame
            Resource meta data with latitude and longitude columns
        shape : tuple
            Desired raster shape in format (number_rows, number_cols)
        vec_dy : np.ndarray
            1D array that represents a (lat, lon) vector
        lat_descending : bool
            Flat to have descending latitudes (this is how the raster would
            appear on the map with north upwards). This option can be changed
            for ease of vertical chunking / indexing.

        Returns
        -------
        raster_index : np.ndarray
            2D array of meta data index values that form a 2D rectangular grid
        """

        iflat = raster_index.flatten()
        lats_raw = meta.loc[iflat, 'latitude'].values
        lons_raw = meta.loc[iflat, 'longitude'].values

        # need to rotate the coordinates to unskew them before sorting lat/lons
        theta = np.arctan2(vec_dy[0], vec_dy[1])
        delta = (np.pi / 2) - theta
        lons = lons_raw * np.cos(delta) - lats_raw * np.sin(delta)
        lats = lats_raw * np.cos(delta) + lons_raw * np.sin(delta)

        # sorting by lat/lons ensures the reshape order
        df = pd.DataFrame({'lats': lats, 'lons': lons}, index=iflat)
        df = df.sort_values(['lons', 'lats'])

        # you need to make sure all the lons in a column are equal otherwise
        # imperfect grid sorting happens
        lons = df['lons'].values.reshape(shape, order='F')
        lons[:] = lons.mean(axis=0)
        df['lons'] = lons.flatten(order='F')
        df = df.sort_values(['lons', 'lats'])

        iflat = df.index.values
        raster_index = iflat.reshape(shape, order='F')

        lons = df['lons'].values.reshape(shape, order='F')
        lats = df['lats'].values.reshape(shape, order='F')

        # make sure lons are ordered correctly
        if (np.diff(lons.mean(axis=0)) < 0).sum() > 0.5 * lons.shape[1]:
            raster_index = raster_index[:, ::-1]

        # make sure lats are ordered correctly
        if (np.diff(lats.mean(axis=1)) < 0).sum() > 0.5 * lats.shape[0]:
            raster_index = raster_index[::-1, :]

        if lat_descending:
            raster_index = raster_index[::-1]
            lats = lats[::-1]

        return raster_index

    @classmethod
    def _get_raster_index(cls, meta, gid_target, vec_dx, vec_dy,
                          shape, lat_descending=True):
        """Get meta data index values that correspond to a 2D rectangular grid
        of the requested shape. This is a hidden compute method that can be
        called iteratively for adaptive sampling.

        Parameters
        ----------
        meta : pd.DataFrame
            Resource meta data with latitude and longitude columns
        gid_target : tuple
            Actual starting coordinates corresponding to a real gid point in
            meta data.
        vector_x : np.ndarray
            1D array of shape (2,) with (delta_latitude, delta_longitude)
            corresponding to the vector for pure positive horizontal movement
            in the meta data
        vector_y : np.ndarray
            1D array of shape (2,) with (delta_latitude, delta_longitude)
            corresponding to the vector for pure positive vertical movement in
            the meta data
        shape : tuple
            Desired raster shape in format (number_rows, number_cols)
        lat_descending : bool
            Flat to have descending latitudes (this is how the raster would
            appear on the map with north upwards). This option can be changed
            for ease of vertical chunking / indexing.

        Returns
        -------
        raster_index : np.ndarray
            2D array of meta data index values that form a 2D rectangular grid
            with latitudes descending from top to bottom and longitudes
            ascending from left to right.
        start_xy : np.ndarray
            1D array of shape (2,) coordinates of the starting search point
        point_x : np.ndarray
            1D array of shape (2,) coordinates of the horizonital search point
        point_y : np.ndarray
            1D array of shape (2,) coordinates of the vertical search point
        end_xy : np.ndarray
            1D array of shape (2,) coordinates of the final search point
        """
        n_vert, n_horiz = shape
        # Set points for origin, horizontal/verical movements, and final
        start_xy = copy.deepcopy(gid_target)
        point_x = copy.deepcopy(gid_target)
        point_y = copy.deepcopy(gid_target)
        end_xy = copy.deepcopy(gid_target)

        # add offsets so bounding box is between grid lines
        start_xy -= (0.5 * (vec_dy + vec_dx))
        point_x += (n_horiz - 0.5) * vec_dx - (0.5 * vec_dy)
        point_y += (n_vert - 0.5) * vec_dy - (0.5 * vec_dx)
        end_xy += (n_vert - 0.5) * vec_dy + (n_horiz - 0.5) * vec_dx

        # slopes of horizontal / vertical vectors
        m_horiz = vec_dx[0] / vec_dx[1]
        m_vert = vec_dy[0] / vec_dy[1]

        # horizontal lines (low, high)
        lin_y_1 = (m_horiz * (meta['longitude'].values - start_xy[1])
                   + start_xy[0])
        lin_y_2 = (m_horiz * (meta['longitude'].values - point_y[1])
                   + point_y[0])

        # vertical lines (left, right)
        lin_x_1 = ((meta['latitude'].values - start_xy[0]) / m_vert
                   + start_xy[1])
        lin_x_2 = ((meta['latitude'].values - point_x[0]) / m_vert
                   + point_x[1])

        # get the mask of the bounding box
        mask = ((meta['latitude'] > lin_y_1)
                & (meta['latitude'] < lin_y_2)
                & (meta['longitude'] > lin_x_1)
                & (meta['longitude'] < lin_x_2))

        if mask.sum() != (n_horiz * n_vert):
            msg = ('Found {} gids but should have found {} by {}. '
                   'Gid target was {}, '
                   'bounding points were calculated to be {} {} {} {},'
                   'and the final found coordinates are: \n{}'
                   .format(mask.sum(), n_horiz, n_vert, gid_target,
                           start_xy, point_x, point_y, end_xy, meta[mask]))
            raise RuntimeError(msg)

        raster_index = meta[mask].index.values
        raster_index = cls._order_raster_index(raster_index, meta,
                                               shape, vec_dy,
                                               lat_descending=lat_descending)

        return raster_index, start_xy, point_x, point_y, end_xy

    def get_raster_index(self, target, shape, meta=None, max_delta=50):
        """Get meta data index values that correspond to a 2D rectangular grid
        of the requested shape starting with the target coordinate in the
        bottom left hand corner. Note that this can break down if a target is
        requested outside of the main grid area.

        Parameters
        ----------
        target : tuple
            Starting coordinate (latitude, longitude) in decimal degrees for
            the bottom left hand corner of the raster grid.
        shape : tuple
            Desired raster shape in format (number_rows, number_cols)
        meta : pd.DataFrame | None
            Optional meta data input with latitude, longitude fields. Default
            is None which extracts self.meta from the resource data.
        max_delta : int
            Optional maximum limit on the raster shape that is retrieved at
            once. If shape is (20, 20) and max_delta=10, the full raseter will
            be retrieved in four chunks of (10, 10). This helps adapt to
            non-regular grids that curve over large distances.

        Returns
        -------
        raster_index : np.ndarray
            2D array of meta data index values that form a 2D rectangular grid
            with latitudes descending from top to bottom and longitudes
            ascending from left to right.
        """

        meta = meta if meta is not None else self.meta

        raster_index = np.zeros(shape, dtype=int)

        next_target = None
        gid_target = self.get_grid_vectors(target, meta=meta)[0]

        # chunk the row (i) and columns (j) rasters
        i_split = int(np.ceil(shape[0] / max_delta))
        j_split = int(np.ceil(shape[1] / max_delta))
        i_chunks = np.array_split(np.arange(shape[0]), i_split)
        j_chunks = np.array_split(np.arange(shape[1]), j_split)

        for ii, i_chunk in enumerate(i_chunks):
            i_slice = slice(i_chunk[0], i_chunk[-1] + 1)
            logger.info('Working on row chunk {} out of {}'
                        .format(ii + 1, len(i_chunks)))

            for jj, j_chunk in enumerate(j_chunks):
                logger.debug('Working on column chunk {} out of {}'
                             .format(jj + 1, len(j_chunks)))
                j_slice = slice(j_chunk[0], j_chunk[-1] + 1)
                temp_shape = (len(i_chunk), len(j_chunk))

                # get the grid vectors using the gid_target from the
                # previous raster chunk
                gid_target, vec_dx, vec_dy, _ = self.get_grid_vectors(
                    gid_target, meta=meta)

                # get the raster using the current grid vectors
                temp, _, point_x, point_y, _ = self._get_raster_index(
                    meta, gid_target, vec_dx, vec_dy, temp_shape,
                    lat_descending=False)

                raster_index[i_slice, j_slice] = temp
                gid_target = point_x + (0.5 * (vec_dx + vec_dy))

                if jj == 0:
                    # save the gid_target for the next row
                    next_target = point_y + (0.5 * (vec_dx + vec_dy))
                elif jj == len(j_chunks) - 1:
                    # use the saved gid_target for the next row
                    gid_target = next_target

        raster_index = raster_index[::-1]

        return raster_index

    @classmethod
    def make_SAM_files(cls, res_h5, gids, out_path, write_time=True,
                       extra_meta_data=None, max_workers=1, n_chunks=36,
                       **kwargs):
        """A performant parallel entry point for making many SAM csv
        files for many gids

        Parameters
        ----------
        res_h5 : str
            Filepath to resource h5 file.
        gids : list | tuple | np.ndarray
            Resource gid(s) of interset
        out_path : str, optional
            Path to save SAM data to in SAM .csv format. A gid index
            "*_{gid}.csv" will be appended to the file path
        write_time : bool
            Flag to write the time columns (Year, Month, Day, Hour, Minute)
        extra_meta_data : dict, optional
            Dictionary that maps the names and values of extra meta
            info. For example, extra_meta_data={'TMY Year': '2020'}
            will add a column 'TMY Year' to the meta data with
            a value of '2020'.
        max_workers : int | None
            Number of parallel workers. None for all workers.
        n_chunks : int
            Number of chunks to split gids into for parallelization
        kwargs : dict
            Internal kwargs for get_SAM_df
        """

        if max_workers == 1:
            with cls(res_h5) as res:
                res.get_SAM_gid(gids, out_path=out_path,
                                write_time=write_time,
                                extra_meta_data=extra_meta_data,
                                **kwargs)
        else:
            msg = 'Bad gids dtype: {}'.format(type(gids))
            assert isinstance(gids, (list, tuple, np.ndarray)), msg
            gid_chunks = np.array_split(np.array(gids), n_chunks)
            with SpawnProcessPool(max_workers=max_workers) as spp:
                for chunk in gid_chunks:
                    spp.submit(cls.make_SAM_files, res_h5, chunk, out_path,
                               write_time=write_time,
                               extra_meta_data=extra_meta_data,
                               max_workers=1, **kwargs)

    def close(self):
        """
        Close res_cls instance
        """
        self._res.close()

    def _get_datasets(self, datasets=None):
        """
        Get datasets to extract, if None extract all datasets

        Parameters
        ----------
        datasets : list | str, optional
            Dataset(s) to extract, by default None

        Returns
        -------
        datasets : list
            Unique set of datasets in alphabetical order
        """
        if datasets is None:
            datasets = self.datasets
        else:
            if isinstance(datasets, str):
                datasets = [datasets]
            else:
                datasets = datasets.copy()

            datasets += ['meta', 'time_index', 'coordinates']

        return sorted(set(datasets))

    def save_subset(self, out_fpath, gids, datasets=None):
        """
        Extract desired datasets for given gids and save to a new
        out_fpath .h5 file

        Parameters
        ----------
        out_fpath : str
            Path to .h5 file to save region datasets to
        gids : list
            List of gids to extract data from and save to .h5
        datasets : str | list, optional
            Dataset(s) to extract from given region and save to out_fpath,
            if None extract all datasets, by default None
        """
        scale_attr = self.resource.SCALE_ATTR
        add_attr = self.resource.ADD_ATTR
        unscale = False

        datasets = self._get_datasets(datasets=datasets)
        with h5py.File(out_fpath, mode='w-') as f_out:
            for k, v in self.global_attrs.items():
                try:
                    f_out.attrs[k] = v
                except Exception as ex:
                    msg = ('Could not transfer global attribute {}: {}\n{}'
                           .format(k, v, ex))
                    warn(msg)

            for dset in datasets:
                if dset in self:
                    ds = self.h5[dset]
                    ds_slice = self._get_ds_slice(dset, gids)
                    data = ResourceDataset.extract(ds, ds_slice,
                                                   scale_attr=scale_attr,
                                                   add_attr=add_attr,
                                                   unscale=unscale)

                    ds_out = f_out.create_dataset(dset,
                                                  shape=data.shape,
                                                  dtype=data.dtype,
                                                  data=data)
                    for k, v in ds.attrs.items():
                        try:
                            ds_out.attrs[k] = v
                        except Exception as ex:
                            msg = ('Could not transfer {} attribute {}: {}\n{}'
                                   .format(dset, k, v, ex))
                            warn(msg)
                else:
                    msg = ("Dataset {} is not available in {} and will "
                           "not be saved to {}".format(dset, self, out_fpath))
                    warn(msg, ResourceWarning)

    def save_region(self, out_fpath, region, datasets=None,
                    region_col='state'):
        """
        Extract desired datasets from desired region and save to a new
        out_fpath .h5 file

        Parameters
        ----------
        out_fpath : str
            Path to .h5 file to save region datasets to
        region : str, optional
            Region to extract all pixels for, by default None
        datasets : str | list, optional
            Dataset(s) to extract from given region and save to out_fpath,
            if None extract all datasets, by default None
        region_col : str, optional
            Region column to search, by default 'state'
        """
        gids = self.region_gids(region, region_col=region_col)

        self.save_subset(out_fpath, gids, datasets=datasets)


class MultiFileResourceX(ResourceX):
    """
    Multi-File resource extraction class
    """

    DEFAULT_RES_CLS = MultiFileResource

    def __init__(self, resource_path, res_cls=None, tree=None,
                 unscale=True, str_decode=True, check_files=False):
        """
        Parameters
        ----------
        resource_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same time index and
            coordinates but can have different datasets.
        res_cls : obj
            Resource class to use to open and access resource data
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        check_files : bool
            Check to ensure files have the same coordinates and time_index
        """
        log_versions(logger)
        res_cls = self.DEFAULT_RES_CLS if res_cls is None else res_cls
        self._res = res_cls(resource_path, unscale=unscale,
                            str_decode=str_decode, check_files=check_files)
        self._dist_thresh = None
        self._tree = tree
        self._i = 0


class MultiYearResourceX(ResourceX):
    """
    Multi Year resource extraction class
    """

    DEFAULT_RES_CLS = Resource

    def __init__(self, resource_path, years=None, tree=None, unscale=True,
                 str_decode=True, res_cls=None, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        resource_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same time index and
            coordinates but can have different datasets.
        years : list, optional
            List of years to access, by default None
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        res_cls : obj
            Resource handler to use to open individual .h5 files
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        log_versions(logger)
        res_cls = self.DEFAULT_RES_CLS if res_cls is None else res_cls
        self._res = MultiYearResource(resource_path, years=years,
                                      unscale=unscale, str_decode=str_decode,
                                      res_cls=res_cls, hsds=hsds,
                                      hsds_kwargs=hsds_kwargs)
        self._dist_thresh = None
        self._tree = tree
        self._i = 0

    def get_means_map(self, ds_name, year=None, region=None,
                      region_col='state', max_workers=None,
                      chunks_per_worker=5):
        """
        Extract given year(s) and compute means

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        year : str | list, optional
            Year(s) to compute means for, by default None
        region : str
            Region to extract all pixels for
        region_col : str
            Region column to search
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5

        Returns
        -------
        ts_map : pandas.DataFrame
            DataFrame of map values
        """
        gids = slice(None)
        if region is not None:
            gids = self.region_gids(region, region_col=region_col)

        if year is None:
            year = slice(None)

        means_map = TemporalStats.run(self.resource.h5_file, ds_name,
                                      sites=gids, statistics='mean',
                                      res_cls=self.resource.__class__,
                                      hsds=self.resource.hsds,
                                      max_workers=max_workers,
                                      chunks_per_worker=chunks_per_worker,
                                      lat_lon_only=True)

        return means_map


class MultiTimeResourceX(ResourceX):
    """
    Resource extraction class for data stored temporaly accross multiple files
    """

    def __init__(self, resource_path, tree=None, unscale=True,
                 str_decode=True, res_cls=None, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        resource_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same time index and
            coordinates but can have different datasets.
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        res_cls : obj
            Resource handler to us to open individual .h5 files
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        log_versions(logger)
        res_cls = self.DEFAULT_RES_CLS if res_cls is None else res_cls
        self._res = MultiTimeResource(resource_path, unscale=unscale,
                                      str_decode=str_decode, res_cls=res_cls,
                                      hsds=hsds, hsds_kwargs=hsds_kwargs)
        self._dist_thresh = None
        self._tree = tree
        self._i = 0


class SolarX(ResourceX):
    """
    Solar Resource extraction class
    """

    DEFAULT_RES_CLS = SolarResource


class NSRDBX(ResourceX):
    """
    NSRDB extraction class
    """

    DEFAULT_RES_CLS = NSRDB


class MultiFileNSRDBX(MultiFileResourceX):
    """
    Multi-File NSRDB extraction class
    """

    DEFAULT_RES_CLS = MultiFileNSRDB


class MultiYearNSRDBX(MultiYearResourceX):
    """
    Multi Year NSRDB extraction class
    """

    DEFAULT_RES_CLS = NSRDB


class MultiTimeNSRDBX(MultiTimeResourceX):
    """
    NSRDB extraction class for data stored temporaly accross multiple files
    """

    DEFAULT_RES_CLS = NSRDB


class WindX(ResourceX):
    """
    Wind Resource extraction class
    """

    DEFAULT_RES_CLS = WindResource

    def get_SAM_gid(self, hub_height, gid, out_path=None, write_time=True,
                    extra_meta_data=None, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given resource gid and hub height

        Parameters
        ----------
        hub_height : int
            Hub height of interest
        gid : int | list
            Resource gid(s) of interset
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        write_time : bool
            Flag to write the time columns (Year, Month, Day, Hour, Minute)
        extra_meta_data : dict, optional
            Dictionary that maps the names and values of extra meta
            info. For example, extra_meta_data={'TMY Year': '2020'}
            will add a column 'TMY Year' to the meta data with
            a value of '2020'.
        kwargs : dict
            Internal kwargs for get_SAM_df:
            - require_wind_dir
            - icing

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        kwargs['height'] = hub_height
        if out_path is not None:
            write_time = False
            kwargs.update({'add_header': True})

        SAM_df = super().get_SAM_gid(gid, out_path=out_path,
                                     write_time=write_time,
                                     extra_meta_data=extra_meta_data,
                                     **kwargs)

        return SAM_df

    def get_SAM_lat_lon(self, hub_height, lat_lon, check_lat_lon=True,
                        out_path=None, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given lat_lon and hub height

        Parameters
        ----------
        hub_height : int
            Hub height of interest
        lat_lon : tuple
            (lat, lon) coordinate of interest
        check_lat_lon : bool, optional
            Flag to check to make sure the requested lat lons are inside the
            resource grid. This is done by comparing with the bounding box of
            the resource coordinates and by ensuring the nearest neighbor
            distance are below the distance threshold to ensure that requested
            lat, lon coordinates are within the resource grid, by default True
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        kwargs : dict
            Internal kwargs for get_SAM_df:
            - require_wind_dir
            - icing

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        gid = self.lat_lon_gid(lat_lon, check_lat_lon=check_lat_lon)
        SAM_df = self.get_SAM_gid(hub_height, gid, out_path=out_path, **kwargs)

        return SAM_df

    @classmethod
    def make_SAM_files(cls, hub_height, res_h5, gids, out_path,
                       write_time=True, extra_meta_data=None, max_workers=1,
                       n_chunks=36, **kwargs):
        """A performant parallel entry point for making many SAM csv
        files for many gids

        Parameters
        ----------
        hub_height : int
            Hub height of interest
        res_h5 : str
            Filepath to resource h5 file.
        gids : list | tuple | np.ndarray
            Resource gid(s) of interset
        out_path : str, optional
            Path to save SAM data to in SAM .csv format. A gid index
            "*_{gid}.csv" will be appended to the file path
        write_time : bool
            Flag to write the time columns (Year, Month, Day, Hour, Minute)
        extra_meta_data : dict, optional
            Dictionary that maps the names and values of extra meta
            info. For example, extra_meta_data={'TMY Year': '2020'}
            will add a column 'TMY Year' to the meta data with
            a value of '2020'.
        max_workers : int | None
            Number of parallel workers. None for all workers.
        n_chunks : int
            Number of chunks to split gids into for parallelization
        kwargs : dict
            Internal kwargs for get_SAM_df
        """
        kwargs['height'] = hub_height
        super().get_SAM_gid(res_h5, gids, out_path, write_time=write_time,
                            extra_meta_data=extra_meta_data,
                            max_workers=max_workers, n_chunks=n_chunks,
                            **kwargs)


class MultiFileWindX(MultiFileResourceX):
    """
    Multi-File Wind Resource extraction class
    """

    DEFAULT_RES_CLS = MultiFileWTK


class MultiYearWindX(MultiYearResourceX):
    """
    Multi Year Wind Resource extraction class
    """

    DEFAULT_RES_CLS = WindResource


class MultiTimeWindX(MultiTimeResourceX):
    """
    Wind resource extraction class for data stored temporaly accross multiple
    files
    """

    DEFAULT_RES_CLS = WindResource


class WaveX(ResourceX):
    """
    Wave data extraction class
    """

    DEFAULT_RES_CLS = WaveResource

    def get_gid_ts(self, ds_name, gid):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        ts : ndarray
            Time-series for given site(s) and dataset
        """
        if ds_name == 'directional_wave_spectrum':
            ts = self[ds_name, :, :, :, gid]
        else:
            ts = self[ds_name, :, gid]

        return ts

    def get_gid_df(self, ds_name, gid):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        df : pandas.DataFrame
            Time-series DataFrame for given site(s) and dataset
        """
        if ds_name == 'directional_wave_spectrum':
            df = self[ds_name, :, :, :, gid]
            index = pd.MultiIndex.from_product(
                [self.time_index, self['frequency'], self['direction']],
                names=['time_index', 'frequency', 'direction'])
            ax1 = np.prod(df.shape[:3])
            ax2 = df.shape[-1] if len(df.shape) == 4 else 1
            df = df.reshape(ax1, ax2)
        else:
            df = self[ds_name, :, gid]
            index = pd.Index(data=self.time_index, name='time_index')

        if isinstance(gid, (int, np.integer)):
            df = pd.DataFrame(df, columns=[gid],
                              index=index)
            df.name = gid
        else:
            df = pd.DataFrame(df, columns=gid,
                              index=index)
            df.name = ds_name

        return df


class MultiYearWaveX(MultiYearResourceX):
    """
    Multi Year Wave extraction class
    """

    DEFAULT_RES_CLS = WaveResource


class MultiTimeWaveX(MultiTimeResourceX):
    """
    Wave resource extraction class for data stored temporaly accross multiple
    files
    """

    DEFAULT_RES_CLS = WaveResource



# -*- coding: utf-8 -*-
"""
Handler utility to parse slicing keys.
"""


def parse_slice(ds_slice):
    """
    Parse dataset slice

    Parameters
    ----------
    ds_slice : tuple | int | slice | list
        Slice to extract from dataset

    Returns
    -------
    ds_slice : tuple
        slice for axis (0, 1)
    """
    if not isinstance(ds_slice, tuple):
        ds_slice = (ds_slice,)

    return ds_slice


def parse_keys(keys):
    """
    Parse keys for complex __getitem__ and __setitem__

    Parameters
    ----------
    keys : string | tuple
        key or key and slice to extract

    Returns
    -------
    key : string
        key to extract
    key_slice : slice | tuple
        Slice or tuple of slices of key to extract
    """
    if isinstance(keys, tuple):
        key = keys[0]
        key_slice = keys[1:]
    else:
        key = keys
        key_slice = (slice(None),)

    return key, key_slice



# -*- coding: utf-8 -*-
"""
Resource Statistics Command Line Interface (CLI)
"""
import click
import logging
import os

from rex.multi_file_resource import MultiFileNSRDB, MultiFileWTK
from rex.renewable_resource import (NSRDB, WaveResource, WindResource)
from rex.resource import Resource
from rex.temporal_stats.temporal_stats import TemporalStats
from rex.utilities.cli_dtypes import INT
from rex.utilities.loggers import init_mult
from rex import __version__

logger = logging.getLogger(__name__)

RES_CLS = {'Resource': Resource,
           'NSRDB': NSRDB,
           'Wind': WindResource,
           'Wave': WaveResource,
           'MultiFileNSRDB': MultiFileNSRDB,
           'MultiFileWTK': MultiFileWTK}


@click.group(invoke_without_command=True, chain=True)
@click.version_option(version=__version__)
@click.option('--resource_path', '-h5', required=True,
              type=click.Path(),
              help=('Path to Resource .h5 files'))
@click.option('--dataset', '-dset', required=True,
              help='Dataset to extract stats for')
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--statistics', '-stats',
              type=click.Choice(['mean', 'median', 'stdev', 'std'],
                                case_sensitive=False), default=['mean'],
              multiple=True, show_default=True,
              help=("Statistics to extract, must be 'mean', 'median', 'std', "
                    "and / or 'stdev'")
              )
@click.option('--max_workers', '-mw', type=INT, default=None,
              show_default=True,
              help=('Number of workers to use, if 1 run in serial, if None use'
                    ' all available cores'))
@click.option('--res_cls', '-res',
              type=click.Choice(['Resource', 'NSRDB', 'Wind', 'Wave',
                                 'MultiFileNSRDB', 'MultiFileWTK'],
                                case_sensitive=False),
              default='Resource', show_default=True,
              help='Resource type')
@click.option('--hsds', '-hsds', is_flag=True,
              help=("Boolean flag to use h5pyd to handle .h5 'files' hosted "
                    "on AWS behind HSDS"))
@click.option('--chunks_per_worker', '-cpw', default=5, type=int,
              show_default=True,
              help='Number of chunks to extract on each worker')
@click.option('--lat_lon_only', '-ll', is_flag=True,
              help='Only append lat, lon coordinates to stats')
@click.option('--log_file', '-log', default=None, type=click.Path(),
              show_default=True,
              help='Path to .log file, if None only log to stdout')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, resource_path, dataset, out_dir, statistics, max_workers,
         res_cls, hsds, chunks_per_worker, lat_lon_only, log_file, verbose):
    """
    TemporalStats Command Line Interface
    """
    ctx.ensure_object(dict)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    name = os.path.splitext(os.path.basename(resource_path))[0]
    out_fpath = '{}_{}.csv'.format(name, dataset)
    out_fpath = os.path.join(out_dir, out_fpath)

    if log_file is not None:
        name = os.path.basename(log_file).split('.')[0]
        log_dir = os.path.dirname(log_file)
    else:
        name = None
        log_dir = None

    init_mult(name, log_dir, [__name__, 'rex'], verbose=verbose)
    logger.info('Computing stats data from {}'.format(resource_path))
    logger.info('Outputs to be stored in: {}'.format(out_dir))

    res_stats = TemporalStats(resource_path, statistics=statistics,
                              res_cls=RES_CLS[res_cls], hsds=hsds)

    if ctx.invoked_subcommand is None:
        all_stats = res_stats.all_stats(
            dataset, max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only)

        res_stats.save_stats(all_stats, out_fpath)
    else:
        ctx.obj['STATS'] = res_stats
        ctx.obj['DSET'] = dataset
        ctx.obj['MAX_WORKERS'] = max_workers
        ctx.obj['CPW'] = chunks_per_worker
        ctx.obj['LL'] = lat_lon_only
        ctx.obj['OUT_PATH'] = out_fpath


@main.command()
@click.pass_context
def full(ctx):
    """
    Compute Stats for full file time-series
    """
    res_stats = ctx.obj['STATS']
    annual_stats = res_stats.full_stats(
        ctx.obj['DSET'], max_workers=ctx.obj["MAX_WORKERS"],
        chunks_per_worker=ctx.obj['CPW'],
        lat_lon_only=ctx.obj['LL'])

    out_fpath = ctx.obj['OUT_PATH'].replace('.csv', '_stats.csv')
    res_stats.save_stats(annual_stats, out_fpath)


@main.command()
@click.pass_context
def monthly(ctx):
    """
    Compute Monthly Stats
    """
    res_stats = ctx.obj['STATS']
    monthly_stats = res_stats.monthly_stats(
        ctx.obj['DSET'], max_workers=ctx.obj["MAX_WORKERS"],
        chunks_per_worker=ctx.obj['CPW'],
        lat_lon_only=ctx.obj['LL'])

    out_fpath = ctx.obj['OUT_PATH'].replace('.csv', '_monthly.csv')
    res_stats.save_stats(monthly_stats, out_fpath)


@main.command()
@click.pass_context
def diurnal(ctx):
    """
    Compute Diurnal Stats
    """
    res_stats = ctx.obj['STATS']
    diurnal_stats = res_stats.diurnal_stats(
        ctx.obj['DSET'], max_workers=ctx.obj["MAX_WORKERS"],
        chunks_per_worker=ctx.obj['CPW'],
        lat_lon_only=ctx.obj['LL'])

    out_fpath = ctx.obj['OUT_PATH'].replace('.csv', '_diurnal.csv')
    res_stats.save_stats(diurnal_stats, out_fpath)


@main.command()
@click.pass_context
def monthly_diurnal(ctx):
    """
    Compute Monthly-Diurnal Stats
    """
    res_stats = ctx.obj['STATS']
    monthly_diurnal_stats = res_stats.monthly_diurnal_stats(
        ctx.obj['DSET'], max_workers=ctx.obj["MAX_WORKERS"],
        chunks_per_worker=ctx.obj['CPW'],
        lat_lon_only=ctx.obj['LL'])

    out_fpath = ctx.obj['OUT_PATH'].replace('.csv', '_monthly_diurnal.csv')
    res_stats.save_stats(monthly_diurnal_stats, out_fpath)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running TemporalStats CLI')
        raise



"""Code for regridding data from one list of coordinates to another"""

import logging
import pickle
import pprint
from functools import cached_property
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, asdict
from datetime import datetime as dt
from typing import Optional

import numpy as np
import pandas as pd
import psutil
from sklearn.neighbors import BallTree


logger = logging.getLogger(__name__)


class _InterpolationMixin:
    """Inverse-weighted distance interpolation logic.

    This mixin class is only intended to be used with classes that
    have the following attributes:

        - self.distances: ndarray of distances from ball tree query
        - self.indices: ndarray of indices from ball tree query
        - self.min_distance: float representing the minimum distance to
                             use for inverse-weighted distances
                             calculation to avoid diving by 0
    """

    @cached_property
    def weights(self):
        """ndarray: Weights used for regridding. """
        return _compute_weights(self.distances, self.min_distance)

    def __call__(self, data):
        """Regrid given spatiotemporal data over entire grid.

        Parameters
        ----------
        data : :obj:`numpy.ndarray` | :obj:`dask.core.array.Array`
            Spatiotemporal data to regrid to target_meta. Data can be
            flattened in the spatial dimension to match the
            `target_meta` or be in a 2D spatial grid, e.g.:
            (spatial, temporal) or (spatial_1, spatial_2, temporal)

        Returns
        -------
        out : :obj:`numpy.ndarray` | :obj:`dask.core.array.Array`
            Flattened regridded spatiotemporal data
            (spatial, temporal)
        """
        if len(data.shape) == 3:
            data = data.reshape((data.shape[0] * data.shape[1], -1))

        msg = "Input data must be 2D (spatial, temporal)"
        assert len(data.shape) == 2, msg

        if hasattr(data, 'vindex'):  # data is Dask array
            vals = data.vindex[self.indices]
        else:
            vals = data[self.indices]

        vals = np.transpose(vals, (2, 0, 1))
        return np.einsum('ijk,jk->ij', vals, self.weights).T


# pylint: disable=attribute-defined-outside-init
@dataclass
class Regridder(_InterpolationMixin):
    """Interpolate from one grid to another using inverse weighted distances.

    This class builds ball tree and runs all queries to create full
    arrays of indices and distances for neighbor points. It computes
    an array of weights used to interpolate from the old grid to the new
    grid.

    Parameters
    ----------
    source_meta : :class:`pandas.DataFrame`
        Set of coordinates for source grid. Must contain "latitude"
        and "longitude" columns representing the coordinates
        (in degrees).
    target_meta : :class:`pandas.DataFrame`
        Set of coordinates for target grid. Must contain "latitude"
        and "longitude" columns representing the coordinates
        (in degrees).
    k_neighbors : int, optional
        Number of nearest neighbors to use for interpolation.
        By default, ``4``.
    n_chunks : int
        Number of spatial chunks to use for tree queries. The total
        number of points in the `target_meta` will be split into
        `n_chunks`, and the points in each chunk will be queried at the
        same time. By default, ``100``.
    max_workers : int, optional
        Max number of workers to use for running all tree queries needed
        to build the full set of indices and distances for each
        `target_meta` coordinate. By default, ``None``, which uses all
        available CPU cores.
    min_distance : float, optional
        Minimum distance to use for inverse-weighted distances
        calculation to avoid diving by 0. By default, ``1e-12``.
    leaf_size : int, optional
        Leaf size for :class:`~sklearn.neighbors.BallTree` instance.
        By default, ``4``.
    """

    source_meta: pd.DataFrame
    target_meta: pd.DataFrame
    k_neighbors: Optional[int] = 4
    n_chunks: Optional[int] = 100
    max_workers: Optional[int] = None
    min_distance: Optional[float] = 1e-12
    leaf_size: Optional[int] = 4

    def __post_init__(self):
        self._tree = None
        self._distances = None
        self._indices = None
        self._weights = None

        fields = pprint.pformat(asdict(self), indent=2)
        logger.info("Initialized `Regridder` with:\n%s", fields)

    @property
    def distances(self):
        """Get distances for all tree queries."""
        if self._distances is None:
            self.init_queries()
        return self._distances

    @property
    def indices(self):
        """Get indices for all tree queries."""
        if self._indices is None:
            self.init_queries()
        return self._indices

    def init_queries(self):
        """Initialize arrays for tree queries and perform all queries"""
        self._indices = [None] * len(self.target_meta)
        self._distances = [None] * len(self.target_meta)
        self.get_all_queries(self.max_workers)

    @property
    def tree(self):
        """Build ball tree from source_meta"""
        if self._tree is None:
            logger.info("Building ball tree for regridding.")
            ll2 = self.source_meta[["latitude", "longitude"]].values
            ll2 = np.radians(ll2)
            self._tree = BallTree(ll2, leaf_size=self.leaf_size,
                                  metric="haversine")
        return self._tree

    def get_all_queries(self, max_workers=None):
        """Query ball tree for all coordinates in the target_meta and store
        results"""

        if max_workers == 1:
            logger.info("Querying all coordinates in serial.")
            self.save_query(slice(None))

        else:
            logger.info("Querying all coordinates in parallel.")
            self._parallel_queries(max_workers=max_workers)
        logger.info("Finished querying all coordinates.")

    def _parallel_queries(self, max_workers=None):
        """Get indices and distances for all points in target_meta, in
        serial"""
        futures = {}
        now = dt.now()
        slices = np.arange(len(self.target_meta))
        slices = np.array_split(slices, min(self.n_chunks, len(slices)))
        slices = [slice(s[0], s[-1] + 1) for s in slices]
        with ThreadPoolExecutor(max_workers=max_workers) as exe:
            for i, s_slice in enumerate(slices):
                future = exe.submit(self.save_query, s_slice=s_slice)
                futures[future] = i
                mem = psutil.virtual_memory()
                msg = ("Query futures submitted: {} out of {}. Current "
                       "memory usage is {:.3f} GB out of {:.3f} GB total."
                       .format(i + 1, len(slices), mem.used / 1e9,
                               mem.total / 1e9))
                logger.info(msg)

            logger.info(f"Submitted all query futures in {dt.now() - now}.")

            for i, future in enumerate(as_completed(futures)):
                idx = futures[future]
                mem = psutil.virtual_memory()
                msg = ("Query futures completed: {} out of {}. Current memory "
                       "usage is {:.3f} GB out of {:.3f} GB total."
                       .format(i + 1, len(futures), mem.used / 1e9,
                               mem.total / 1e9))
                logger.info(msg)
                try:
                    future.result()
                except Exception as e:
                    msg = ("Failed to query coordinate chunk with index={}"
                           .format(idx))
                    logger.exception(msg)
                    raise RuntimeError(msg) from e

    def save_query(self, s_slice):
        """Save tree query for coordinates specified by given spatial slice"""
        out = self.tree.query(self.get_spatial_chunk(s_slice),
                              k=self.k_neighbors)
        self.distances[s_slice] = out[0]
        self.indices[s_slice] = out[1]

    def get_spatial_chunk(self, s_slice):
        """Get list of coordinates in target_meta specified by the given
        spatial slice

        Parameters
        ----------
        s_slice : slice
            Slice specifying which spatial indices in the target grid
            should be selected. This selects `n_points` from the target
            grid.

        Returns
        -------
        ndarray
            Array of `n_points` in `target_meta` selected by `s_slice`.
        """
        out = self.target_meta.iloc[s_slice][["latitude", "longitude"]].values
        return np.radians(out)

    @classmethod
    def run(cls, source_meta, target_meta, source_data, k_neighbors=4,
            n_chunks=100, max_workers=None, min_distance=1e-12,
            leaf_size=4):
        """Regrid data using inverse distance weighting.

        Parameters
        ----------
        source_meta : :class:`pandas.DataFrame`
            Set of coordinates for source grid. Must contain "latitude"
            and "longitude" columns representing the coordinates
            (in degrees).
        target_meta : :class:`pandas.DataFrame`
            Set of coordinates for target grid. Must contain "latitude"
            and "longitude" columns representing the coordinates
            (in degrees).
        source_data : ndarray
            Spatiotemporal data to regrid to `target_meta` coordinate
            grid. Data can be flattened in the spatial dimension to
            match the `target_meta` or be in a 2D spatial grid, e.g.:
            (spatial, temporal) or (spatial_1, spatial_2, temporal)
        leaf_size : int, optional
            Leaf size for :class:`~sklearn.neighbors.BallTree` instance.
            By default, ``4``.
        k_neighbors : int, optional
            Number of nearest neighbors to use for interpolation.
            By default, ``4``.
        n_chunks : int
            Number of spatial chunks to use for tree queries. The total
            number of points in the `target_meta` will be split into
            `n_chunks`, and the points in each chunk will be queried at
            the same time. By default, ``100``.
        max_workers : int, optional
            Max number of workers to use for running all tree queries
            needed to build the full set of indices and distances for
            each `target_meta` coordinate. By default, ``None``, which
            uses all available CPU cores.
        min_distance : float, optional
            Minimum distance to use for inverse-weighted distances
            calculation to avoid diving by 0. By default, ``1e-12``.
        """
        regridder = cls(source_meta=source_meta, target_meta=target_meta,
                        leaf_size=leaf_size, k_neighbors=k_neighbors,
                        n_chunks=n_chunks, max_workers=max_workers,
                        min_distance=min_distance)
        regridder.get_all_queries(max_workers)
        return regridder(source_data)


class CachedRegridder(_InterpolationMixin):
    """Interpolate from one grid to another using cached dists and inds."""

    def __init__(self, cache_pattern, min_distance=1e-12):
        """

        Parameters
        ----------
        cache_pattern : str
            Filepath pattern for cached distances and indices to load.
            Should be of the form ``'./{array_name}.pkl'`` where
            `array_name` will internally be replaced with either
            ``'distances'`` or ``'indices'``.'
        min_distance : float, optional
            Minimum distance to use for inverse-weighted distances
            calculation to avoid diving by 0. By default, ``1e-12``.
        """
        self.distances, self.indices = self.load_cache(cache_pattern)
        self.min_distance = min_distance

    @staticmethod
    def load_cache(cache_pattern):
        """Load cached indices and distances from ball tree query.

        Parameters
        ----------
        cache_pattern : str
            Filepath pattern for cached distances and indices to load.
            Should be of the form ``'./{array_name}.pkl'`` where
            `array_name` will internally be replaced with either
            ``'distances'`` or ``'indices'``.

        Returns
        -------
        distances, indices : ndarray
            Arrays of distances and indices output by the ball tree.
        """
        distance_file = cache_pattern.format(array_name='distances')
        index_file = cache_pattern.format(array_name='indices')

        with open(distance_file, 'rb') as f:
            distances = pickle.load(f)
        with open(index_file, 'rb') as f:
            indices = pickle.load(f)

        logger.info('Loaded cache files: %s, %s', distance_file, index_file)
        return distances, indices

    @classmethod
    def build_cache(cls, cache_pattern, *args, **kwargs):
        """Cache distances and indices from ball tree query.

        Parameters
        ----------
        cache_pattern : str
            Filepath pattern used to cache distances and indices.
            Should be of the form ``'./{array_name}.pkl'`` where
            `array_name` will internally be replaced with either
            ``'distances'`` or ``'indices'``.
        *args, **kwargs
            Arguments followed by keyword arguments that can be used to
            initialize :class:`Regridder`. The ``Regridder`` instance
            will generate the distance and index arrays to be cached.
        """
        distance_file = cache_pattern.format(array_name='distances')
        index_file = cache_pattern.format(array_name='indices')

        regridder = Regridder(*args, **kwargs)

        with open(distance_file, 'wb') as f:
            pickle.dump(regridder.distances, f, protocol=4)
        with open(index_file, 'wb') as f:
            pickle.dump(regridder.indices, f, protocol=4)
        logger.info('Saved cache files: %s, %s', distance_file, index_file)


def _compute_weights(distances, min_distance):
    """Compute inverse weights from distance values. """
    dists = np.array(distances, dtype=np.float32)
    mask = dists < min_distance
    dists[mask] = min_distance
    if mask.sum() > 0:
        logger.info("%d of %d neighbor distances are within %.5f.",
                    np.sum(mask), np.prod(mask.shape),
                    min_distance)
    weights = 1 / dists
    weights[mask.any(axis=1), :] = (np.eye(1, dists.shape[1])
                                    .flatten())
    return weights / np.sum(weights, axis=-1)[:, None]
